from __future__ import division
from numpy import pi, sqrt, sin, cos, tan, real, imag
import numpy as np

from . import logger
from .const import g


# supported fiber geometries
FIBER_TYPES = [
    'Round',
    'Ribbon',
    'Tapered',
]


def generate_symbolic_tfs(stages=4):
    import sympy as sp

    # construct quad pendulum equation of motion matrix
    ksyms = sp.numbered_symbols('k')
    msyms = sp.numbered_symbols('m')
    w = sp.symbols('w')
    k = [next(ksyms) for n in range(stages)]
    m = [next(msyms) for n in range(stages)]
    A = sp.zeros(stages)
    for n in range(stages-1):
        # mass and restoring forces (diagonal elements)
        A[n, n] = k[n] + k[n+1] - m[n] * w**2
        # couplings to stages above and below
        A[n, n+1] = -k[n+1]
        A[n+1, n] = -k[n+1]
    # mass and restoring force of bottom stage
    A[-1, -1] = k[-1] - m[-1] * w**2

    # want TM equations of motion, so index 4
    b = sp.zeros(stages, 1)
    b[-1] = 1

    # solve linear system
    xsyms = sp.numbered_symbols('x')
    x = [next(xsyms) for n in range(stages)]
    ans = sp.linsolve((A, b), x)
    return ans


def tst_force_to_tst_displ(k, m, f):
    """transfer function for quad pendulum

    """
    k0, k1, k2, k3 = k
    m0, m1, m2, m3 = m
    w = 2*pi*f
    X3 = (k2**2*(k0 + k1 - m0*w**2) + (k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2))/(-k3**2*(k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2)) + (k3 - m3*w**2)*(k2**2*(k0 + k1 - m0*w**2) - (-k1**2 + (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2)))
    return X3


def top_displ_to_tst_displ(k, m, f):
    """transfer function for quad pendulum

    """
    k0, k1, k2, k3 = k
    m0, m1, m2, m3 = m
    w = 2*pi*f
    X0 = k1*k2*k3/(k3**2*(k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2)) - (k3 - m3*w**2)*(k2**2*(k0 + k1 - m0*w**2) + (k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2)))
    return X0 * k0


def getJointParams(sus, n):
    """Return relevant material properties from SUS struct

    Parameters are returned as a pair of tuples,
    (isLower, Temperature, WireMaterialStruct, BladeMaterialStruct),
    corresponding to the upper and lower joints of stage n.

    """
    # Here the suspension stage list is reversed so that the last stage
    # in the list is the test mass
    stages = sus.Stage[::-1]
    last_stage = len(stages) - 1
    stage = stages[n]
    if 'Temp' in stage:
        Temp = stage.Temp
    else:
        Temp = sus.Temp

    TempLower = Temp
    if n == 0:
        # For the top stage: if sus.Temp exists, interpret it as the
        # upper joint temperature.  Otherwise, assume upper and lower
        # joints of the top stage have the same temperature.
        if 'Temp' in sus:
            TempUpper = sus.Temp
        else:
            TempUpper = Temp
    else:
        # Below the top stage: assume that the upper joint has the same
        # temperature as the nearby lower joint of the preceding stage.
        if 'Temp' in stages[n-1]:
            TempUpper = stages[n-1].Temp
        else:
            TempUpper = sus.Temp

    ##############################
    # material parameters

    # support wire
    # upper and lower joint material properties may be different
    # due to temperature gradient
    if 'WireMaterialUpper' in stage and 'WireMaterialLower' in stage:
        wireMatUpper = stage.WireMaterialUpper
        wireMatLower = stage.WireMaterialLower
    elif 'WireMaterial' in stage:
        wireMatUpper = stage.WireMaterial
        wireMatLower = stage.WireMaterial
    elif n == last_stage:
        wireMatUpper = 'Silica'
        wireMatLower = 'Silica'
    else:
        wireMatUpper = 'C70Steel'
        wireMatLower = 'C70Steel'

    WireMaterialUpper = sus[wireMatUpper]
    WireMaterialLower = sus[wireMatLower]
    logger.debug('stage {} wires: {}, {}'.format(n, wireMatUpper, wireMatLower))

    # support blade (upper joint only)
    if 'BladeMaterial' in stage:
        bladeMat = stage.BladeMaterial
    else:
        bladeMat = 'MaragingSteel'

    BladeMaterial = sus[bladeMat]

    logger.debug('stage {} blades: {}'.format(n, bladeMat))

    ULparams = ((0, TempUpper, WireMaterialUpper, BladeMaterial),
                (1, TempLower, WireMaterialLower, None))
    return ULparams


def wireGeometry(r, N, RibbonThickness=None, TaperedEndRadius=None, **kwargs):
    """Compute wire geometry-dependent factors

    r is the wire radius, or ribbon width.
    N is the number of wires or ribbons
    RibbonThickness must be set when ribbons are used, or
    TaperedEndRadius when tapered fibers are used.
    Other kwargs are ignored.

    Returns cross-sectional areas (central and end) and moment of inertia,
    and modified surface to volume ratios (vertical and horizontal).

    """
    # Usual case: round wire/fiber
    xsect = pi * r**2  # cross-sectional area
    xsectEnd = xsect   # only differs for tapered fiber
    # cross-sectional moment of inertia
    # Ref: Gretarsson et al, Phys. Lett. A 270 (2000) 108-114 Eq.(2)
    xII = r**4 * pi / 4  # x-sectional moment of inertia
    # surface to volume ratio, vertical
    # Surface = 2*pi*r*h
    # Volume  = pi*r^2*h
    # Geometrical factor \mu is unity (corresponding to uniform strain energy)
    mu_v = 2 / r
    # surface to volume ratio, horizontal
    # Ref: Gretarsson and Harry, Rev. Sci. Inst. 70 (1999) 4081-4087
    # Appendix (A8)
    # There is a geometrical factor \mu to calculate the ratio of the
    # energies between the ones stored in surface and the bulk.
    # For a transversely oscillating fiber of circular cross section, this
    # factor is \mu = 2 as shown in (A11)
    mu_h = mu_v * 2

    # Special case: ribbon
    if RibbonThickness is not None:
        W = r
        t = RibbonThickness
        xsect = W * t        # cross-sectional area
        xsectEnd = xsect     # only differs for tapered fiber
        xII = (W * t**3)/12  # x-sectional moment of inertia
        # surface to volume ratio, vertical
        # Surface = 2*(W+t)*h
        # Volume  = W*t*h
        mu_v = 2 * (W + t) / (W * t)
        # surface to volume ratio, horizontal
        # Ref: Gretarsson et al, Phys. Lett. A 270 (2000) 108-114
        # There is a geometrical factor \mu (same as above).
        # This factor is shown in Eq.(6)
        mu_h = mu_v * (3 * N * W + t) / (N * W + t)

    # Special case: tapered
    elif TaperedEndRadius is not None:
        r_end = TaperedEndRadius
        xsectEnd = pi * r_end**2  # cross-sectional area (for delta_h)
        xII = pi * r_end**4 / 4   # x-sectional moment of inertia
        mu_h = 4 / r_end          # surface to volume ratio, horizontal
        # mu_v is unchanged (assumes the middle part of the fiber dominates)

    return (xsect, xsectEnd, xII, mu_v, mu_h)


def wireTELoss(w, tension, xsectEnd, xII, Temp, alpha, beta, rho, C, K, Y,
               RibbonThickness=None, **kwargs):
    """Thermoelastic calculations for wires

    Repeated for upper and lower joint of each stage.
    w = angular frequency
    tension = weight supported per wire
    xsectEnd = cross sectional area of wire end
    xII = cross sectional moment of inertia
    Temp = temperature
    alpha = coeff of thermal expansion
    beta = temp dependence of Young's modulus
    rho = mass density
    C = heat capacity
    K = thermal conductivity W/(m K)
    Y = Young's modulus
    RibbonThickness must be set when ribbons are used
    Other kwargs are ignored

    Returns the loss angle associated with thermoelastic damping
    (wire horizontal)

    """
    # horizontal TE time constant, wires
    # The constant 7.37e-2 is 1/(4*q0^2) from eq 12, C. Zener 10.1103/PhysRev.53.90 (1938)
    tau = 7.37e-2 * 4 * (rho * C * xsectEnd) / (pi * K)

    # deal with ribbon geometry
    if RibbonThickness is not None:
        t = RibbonThickness
        tau = (rho * C * t**2) / (K * pi**2)

    # delta: TE factor
    # The first term expresses the cancellation of the CTE and dY/dT
    # beta = (dY/dT)/Y
    # Refs:
    # Cagnoli G and Willems P 2002 Phys. Rev. B 65 174111
    # Cumming A V et al 2009 Class. Quantum Grav. 26 215012
    # horizontal delta, wires
    delta = (alpha - tension * beta / (xsectEnd * Y))**2 * Y * Temp / (rho * C)

    phi_TE = delta * tau * w / (1 + w**2 * tau**2)
    return phi_TE


def bladeTELoss(w, t, Temp, alpha, beta, rho, C, K, Y):
    """Thermoelastic calculations for blades

    Invoked for upper joint only (there is no lower blade)
    w = angular frequency
    t = blade thickness
    Temp = temperature
    alpha = coeff of thermal expansion
    beta = temp dependence of Young's modulus
    rho = mass density
    C = heat capacity
    K = thermal conductivity W/(m K)
    Y = Young's modulus

    Returns the loss angle associated with thermoelastic damping
    (blade vertical)

    """
    # vertical TE time constant, blades
    tau = (rho * C * t**2) / (K * pi**2)

    # vertical delta, blades
    # Here the TE cancellation is ignored
    delta = Y * alpha**2 * Temp / (rho * C)

    phi_TE = delta * tau * w / (1 + w**2 * tau**2)
    return phi_TE


##############################
# Ref "GG"
#   Suspensions thermal noise in the LIGO gravitational wave detector
#   Gabriela Gonzalez, https://doi.org/10.1088/0264-9381/17/21/305
# Ref "GS"
#   Brownian motion of a mass suspended by an anelastic wire
#   Gonzalez & Saulson, https://doi.org/10.1121/1.410467
#
# Note:
#  I in GG = xII
#  rho in GG = rho * xsect
#  delta in GG = d_bend
#  T in GG = tension
##############################

def continuumWireKh(w, N, length, tension, xsect, xII, rho, Y, phi):
    """Horizontal spring constant, including violin modes

    w = angular frequency
    N = number of wires
    length = wire length
    tension = weight supported per wire
    xsect = wire cross sectional area
    xII = cross sectional moment of inertia
    rho = mass density
    Y = Young's modulus
    phi = loss angle

    Returns the spring constant (wire horizontal)
    """
    Y = Y * (1 + 1j * phi)  # complex Young's modulus

    # simplification factors for later calculations
    # Note: a previous version of this suspension thermal noise
    # calculation used a factor "simp3" which was an incorrect
    # approximation for k_e (see GS eq 10).  Here we approximate k_e
    # as 1/delta.  This is equivalent to assuming delta << 1/k,
    # i.e. the bending length of the fiber is much shorter than the
    # 'ideal string' wavelength.
    k = sqrt(rho * xsect / tension) * w
    d_bend = sqrt(Y * xII / tension)  # complex d_bend
    dk = k * d_bend                   # dk << 1 was assumed
    coskl = cos(k * length)
    sinkl = sin(k * length)

    # numerator, horiz spring constant
    #   numerator of K_xx in eq 9 of GG
    #     = T k (cos(k L) + k delta sin(k L))
    #   for w -> 0, this reduces to N_w * T * k
    khnum = N * tension * k * (coskl + dk * sinkl)

    # denominator, horiz spring constant
    #   D after equation 8 in GG
    #   D = sin(k L) - 2 k delta cos(k L)
    #   for w -> 0, this reduces to k (L - 2 delta)
    khden = sinkl - 2 * dk * coskl

    return khnum/khden


def continuumWireKv(w, N, length, xsect, xsectEnd, rho, Y, phi,
                    TaperedEndLength=None, **kwargs):
    """Vertical spring constant, including bounce mode.

    w = angular frequency
    N = number of wires
    length = wire length
    xsect = wire cross sectional area
    xsectEnd = cross sectional area of wire end
    rho = mass density
    Y = Young's modulus
    phi = loss angle
    TaperedEndLength must be set when tapered fibers are used
    Other kwargs are ignored

    Returns the spring constant (wire vertical)
    """
    Y = Y * (1 + 1j * phi)  # complex Young's modulus
    k = sqrt(rho / Y) * w

    kv = N * xsect * Y * k / tan(k * length)
    # deal with tapered geometry
    if TaperedEndLength is not None:
        l_end = TaperedEndLength
        l_mid = length - 2*l_end
        kv_mid = N * xsect * Y * k / tan(k * l_mid)
        kv_end = N * xsectEnd * Y * k / tan(k * l_end)
        kv = 1/(2/kv_end + 1/kv_mid)

    return kv


def suspQuad(f, sus):
    """Suspension for quadruple pendulum

    `f` is frequency vector, `sus` is suspension model.
    Violin modes are included for the bottom stage.

      sus.FiberType should be: 0=round, 1=ribbons.

    hForce, vForce = transfer functions from the force on the TM to TM
    motion.  These should have the correct losses for the mechanical
    system such that the thermal noise is:

    dxdF = force on TM along beam line to position of TM along beam line
         = hForce + theta^2 * vForce
         = admittance / (i * w)

    where theta = sus.VHCoupling.theta.

    Since this is just suspension thermal noise, the TM internal modes
    and coating properties should not be included.

    hTable, vTable = TFs from support motion to TM motion

    Ah = horizontal equations of motion
    Av = vertical equations of motion

    Adapted from code by Morag Casey (Matlab) and Geppo Cagnoli
    (Maple).  Modification for the different temperatures between the
    stages by K.Arai.

    """
    w = 2 * pi * f

    # NOTE: For historical reasons, the IFO struct numbers stages from the
    # bottom up (0 is the TM), while this code numbers them from the top down
    # (3 is the TM).  Here the suspension stage list is reversed so that the
    # last stage in the list is the test mass
    stages = sus.Stage[::-1]
    last_stage = len(stages) - 1

    # bottom stage fiber type
    FiberType = sus.FiberType
    assert FiberType in FIBER_TYPES

    ##############################
    # Compute cumulative weight of suspension
    masses = np.array([stage.Mass for stage in stages])
    # weight support by lower stages
    Mgs = g * np.flipud(np.cumsum(np.flipud(masses)))

    ##############################
    # Complex spring constants
    khr = np.zeros([len(stages), len(w)])
    khi = np.zeros([len(stages), 2, len(w)])
    kvr = np.zeros([len(stages), len(w)])
    kvi = np.zeros([len(stages), 2, len(w)])

    for n, stage in enumerate(stages):

        ##############################
        # main suspension parameters
        Mg = Mgs[n]
        length = stage.Length

        # Correction for the pendulum restoring force
        kh0 = Mg / length   # N/m, horiz. spring constant, stage n

        kv0_blade = stage.K    # N/m, vert. spring constant (from blade)

        if not np.isnan(stage.WireRadius):
            r_w = stage.WireRadius
        elif FiberType == 'Ribbon':
            r_w = sus.Ribbon.Width
        else:
            r_w = sus.Fiber.Radius

        # blade thickness
        t_b = stage.Blade
        # number of support wires
        N_w = stage.NWires
        tension = Mg / N_w

        # wire geometry
        wireShape = {}
        if n == last_stage and FiberType == 'Ribbon':
            wireShape['RibbonThickness'] = sus.Ribbon.Thickness
        elif n == last_stage and FiberType == 'Tapered':
            wireShape['TaperedEndRadius'] = sus.Fiber.EndRadius
            wireShape['TaperedEndLength'] = sus.Fiber.EndLength
        xsect, xsectEnd, xII, mu_v, mu_h = wireGeometry(r_w, N_w, **wireShape)

        jointParams = getJointParams(sus, n)
        for (isLower, Temp, wireMat, bladeMat) in jointParams:
            # wire parameters
            alpha_w = wireMat.Alpha  # coeff of thermal expansion
            beta_w = wireMat.dlnEdT  # temp dependence of Young's modulus
            rho_w = wireMat.Rho      # mass density
            C_w = wireMat.C          # heat capacity
            K_w = wireMat.K          # thermal conductivity W/(m K)
            Y_w = wireMat.Y          # Young's modulus
            phi_w = wireMat.Phi      # loss angle

            # surface loss dissipation depth
            # About the surface loss depth see Ref:
            # A. M. Gretarsson and G. M. Harry
            # Rev. Sci. Instrum. 70 (1999) P.4081-4087
            if 'Dissdepth' in wireMat:
                ds_w = wireMat.Dissdepth
            else:
                ds_w = 0             # ignore surface effects

            if not isLower:  # blade parameters (there is no lower blade)
                alpha_b = bladeMat.Alpha # coeff of thermal expansion
                beta_b = bladeMat.dlnEdT # temp dependence of Young's modulus
                rho_b = bladeMat.Rho     # mass density
                C_b = bladeMat.C         # heat capacity
                K_b = bladeMat.K         # thermal conductivity W/(m K)
                Y_b = bladeMat.Y         # Young's modulus
                phi_b = bladeMat.Phi     # loss angle

            # bending length, and dilution factors
            if 'Dilution' in stage and not np.isnan(stage.Dilution):
                # if a specific dilution factor is given, use it
                dil = stage.Dilution
            else:
                # if not, calculate
                # This formula is for the effective dilution in the GW band
                # (as discussed in ref GG section 2.5)
                d_bend = sqrt(Y_w * xII / tension)
                dil = length / d_bend

            ##############################
            # Loss Calculations for wires and blades
            # Track losses from the upper and lower wire joints separately
            # The factor of 2 comes from that the loss was calculated
            # for both upper and lower joints
            # Blades are always attached at the top, there is no "lower" blade
            # The final stage is handled specially and the spring constant
            # calculation below is not used

            # horizontal loss factor for round wires
            # phi_w: bulk loss (brownian motion)
            # (mu_h*ds_w): surface loss (brownian motion)
            # nominally this term becomes zero for steel wires as ds_w is zero
            # The last term is for TE loss
            phih_TE = wireTELoss(w, tension, xsectEnd, xII, Temp,
                                 alpha_w, beta_w, rho_w, C_w, K_w, Y_w,
                                 **wireShape)
            phih = phi_w * (1 + mu_h * ds_w) + phih_TE

            # complex spring constant, horizontal
            khr[n, :] = kh0
            khi[n, isLower, :] = kh0 * phih / dil / 2

            if not isLower:
                # vertical loss factor, blades
                # The bulk loss term and the TE loss term
                # wire loss and stiffness are neglected here
                phiv_TE = bladeTELoss(w, t_b, Temp,
                                      alpha_b, beta_b, rho_b, C_b, K_b, Y_b)
                phiv_blade = phi_b + phiv_TE

                # complex spring constant, vertical
                kvr[n, :] = kv0_blade
                kvi[n, 0, :] = kv0_blade * phiv_blade


        if n == last_stage:
            # loss factor, vertical (no blades)
            # wire vertical thermoelastic loss is negligible (Saulson 1990)
            phiv = phi_w * (1 + mu_v * ds_w)

            kv4 = continuumWireKv(w, N_w, length, xsect, xsectEnd,
                                  rho_w, Y_w, phiv, **wireShape)
            kh4 = continuumWireKh(w, N_w, length, tension, xsect, xII,
                                  rho_w, Y_w, phih)

            # if blades present, combine two vertical springs
            if not np.isnan(kv0_blade):
                kv4wire = kv4
                kv4blade = kvr[n, :] + 1j*kvi[n, 0, :]
                kv4 = 1/(1/kv4wire + 1/kv4blade)

            kvr[n, :] = real(kv4)
            kvi[n, 0, :] = imag(kv4)
            kvi[n, 1, :] = 0
            khr[n, :] = real(kh4)
            khi[n, 0, :] = imag(kh4) / 2
            khi[n, 1, :] = imag(kh4) / 2


    ###############################################################
    # Equations of motion for the system
    ###############################################################

    # Calculate horizontal/vertical TFs turning on the loss of each
    # joint one by one, for the thermal noise calculation
    hForce = np.zeros([2*len(stages), len(w)], dtype=complex)
    vForce = np.zeros([2*len(stages), len(w)], dtype=complex)

    for m in range(2*len(stages)):
        # turn on only the loss of the current joint
        n = int(m/2)  # stage number
        isLower = m % 2
        stage_selection = np.zeros((len(stages), 1))
        stage_selection[n] = 1

        # horizontal
        # only the imaginary part due to the specified joint is used.
        k = khr + 1j*khi[:,isLower,:]*stage_selection
        # calculate TFs
        hForce[m,:] = tst_force_to_tst_displ(k, masses, f)

        # vertical
        # only the imaginary part due to the specified joint is used
        k = kvr + 1j*kvi[:,isLower,:]*stage_selection
        # calculate TFs
        vForce[m,:] = tst_force_to_tst_displ(k, masses, f)

    # Calculate horizontal/vertical TFs with all losses on,
    # for the vibration isolation calculation
    k = kvr + 1j*(kvi[:,0,:] + kvi[:,1,:])
    vTable = top_displ_to_tst_displ(k, masses, f)

    k = khr + 1j*(khi[:,0,:] + khi[:,1,:])
    hTable = top_displ_to_tst_displ(k, masses, f)

    # test mass susceptibility for radiation pressure calculations
    tst_suscept = tst_force_to_tst_displ(k, masses, f)

    return hForce, vForce, hTable, vTable, tst_suscept



"""
"""
import pytest

from gwinc import load_budget


@pytest.mark.fast
@pytest.mark.logic
def test_inherit_load(pprint, tpath_join, fpath_join):
    fpath = fpath_join('Aplus_mod.yaml')
    B_inherit = load_budget(fpath)
    B_orig = load_budget('Aplus')

    pprint(B_inherit.ifo)
    pprint("ACTUAL TEST")
    pprint(B_inherit.ifo.diff(B_orig.ifo))

    assert(
        sorted(B_inherit.ifo.diff(B_orig.ifo))
        == sorted([
            ('Suspension.Stage[3].Mass', 30, 22.1),
            ('Squeezer.AmplitudedB', 14, 12),
            ('Squeezer.InjectionLoss', 0.02, 0.05)])
    )

    fpath2 = fpath_join('Aplus_mod2.yaml')
    B_inherit2 = load_budget(fpath2)
    pprint(B_inherit2.ifo.diff(B_orig.ifo))
    assert(
        sorted(B_inherit2.ifo.diff(B_orig.ifo))
        == sorted([
            ('Suspension.Stage[2].Mass', 30, 21.8),
            ('Suspension.Stage[3].Mass', 30, 22.1),
            ('Squeezer.InjectionLoss', 0.02, 0.05)
        ])
    )


@pytest.mark.fast
@pytest.mark.logic
def test_inherit_fail(pprint, tpath_join, fpath_join):
    """
    This test shows that the interim logic that Struct.from_file checks for and fails when it gets "+inherit" keys.
    Those will be allowable at a later time.
    """
    fpath2 = fpath_join('inherit_fail.yaml')

    with pytest.raises(RuntimeError):
        B_inherit2 = load_budget(fpath2)



'''Functions to calculate residual gas noise

'''
from __future__ import division
from numpy import sqrt, log, pi

from .. import const
from .. import Struct


def residual_gas_scattering_arm(f, ifo, cavity, species):
    """Residual gas noise strain spectrum due to scattering from one arm

    Noise caused by the passage of residual gas molecules through the
    laser beams in one arm cavity due to scattering.

    :f: frequency array in Hz
    :ifo: gwinc IFO structure
    :cavity: arm cavity structure
    :species: molecular species structure

    :returns: arm strain noise power spectrum at :f:

    The method used here is presented by Rainer Weiss, Micheal
    E. Zucker, and Stanley E. Whitcomb in their paper Optical
    Pathlength Noise in Sensitive Interferometers Due to Residual Gas.

    Added to Bench by Zhigang Pan, Summer 2006
    Cleaned up by PF, Apr 07
    Eliminated numerical integration and substituted first order
    expansion of exp, to speed it up.

    """
    L = ifo.Infrastructure.Length
    kT = ifo.Infrastructure.Temp * const.kB
    P = species.BeamtubePressure
    M = species.mass
    alpha = species.polarizability

    rho = P / (kT)                   # number density of Gas
    v0 = sqrt(2*kT / M)              # mean speed of Gas

    waist = cavity.w0                # Gaussian beam waist size
    zr = cavity.zr                   # Rayleigh range
    z1 = -cavity.zBeam_ITM           # location of ITM relative to the waist
    z2 = cavity.zBeam_ETM            # location of ETM relative to the waist

    # The exponential of Eq. 1 of P940008 is expanded to first order; this
    # can be integrated analytically
    zint = log(z2 + sqrt(z2**2 + zr**2)) - log(z1 + sqrt(z1**2 + zr**2))
    zint = zint * zr/waist
    zint = zint - 2*pi*L*f/v0
    # optical path length for one arm
    zint = zint*((4*rho*(2*pi*alpha)**2)/v0)
    # eliminate any negative values due to first order approx.
    zint[zint < 0] = 0

    return zint


def residual_gas_damping_test_mass(f, ifo, species, sustf, squeezed_film):
    """Noise due to residual gas damping for one test mass

    :f: frequency array in Hz
    :ifo: gwinc IFO structure
    :species: molecular species structure
    :sustf: suspension transfer function structure
    :squeezed_film: squeezed film damping structure

    :returns: displacement noise
    """
    sus = ifo.Suspension
    if 'Temp' in sus.Stage[0]:
        kT = sus.Stage[0].Temp * const.kB
    else:
        kT = sus.Temp * const.kB

    mass = species.mass
    radius = ifo.Materials.MassRadius
    thickness = ifo.Materials.MassThickness
    thermal_vel = sqrt(kT/mass)  # thermal velocity

    # pressure in the test mass chambers; possibly different from the pressure
    # in the arms due to outgassing near the test mass
    pressure = species.ChamberPressure

    # infinite volume viscous damping coefficient for a cylinder
    # table 1 of https://doi.org/10.1016/j.physleta.2010.06.041
    beta_inf = pi * radius**2 * pressure/thermal_vel
    beta_inf *= sqrt(8/pi) * (1 + thickness/(2*radius) + pi/4)

    force_noise = 4 * kT * beta_inf

    # add squeezed film damping if necessary as parametrized by
    # Eq (5) of http://dx.doi.org/10.1103/PhysRevD.84.063007
    if squeezed_film.keys():
        # the excess force noise and diffusion time are specified directly
        if 'ExcessDamping' in squeezed_film:
            # ExcessDamping is the ratio of the total gas damping noise at DC
            # to damping in the infinite volume limit (in amplitude)
            DeltaS0 = (squeezed_film.ExcessDamping**2 - 1) * force_noise
            if DeltaS0 < 0:
                raise ValueError('ExcessDamping must be > 1')

            try:
                diffusion_time = squeezed_film.DiffusionTime
            except AttributeError:
                msg = 'If excess residual gas damping is given a diffusion ' \
                    + 'time must be specified as well'
                raise ValueError(msg)

        # if a gap between the test mass and another object is specified
        # use the approximate model of section IIIA and B
        elif 'gap' in squeezed_film:
            gap = squeezed_film.gap

            # Eq (14)
            diffusion_time = sqrt(pi/2) * radius**2 / (gap * thermal_vel)
            diffusion_time /= log(1 + (radius/gap)**2)

            # Eq (11) factoring out the low pass cutoff as in (5)
            DeltaS0 = 4 * kT * pi*radius**2 * pressure * diffusion_time / gap

        else:
            raise ValueError('Must specify either excess damping or a gap')

        # Eq (5)
        force_noise += DeltaS0 / (1 + (2*pi*f*diffusion_time)**2)

    # convert force to displacement noise using the suspension susceptibility
    noise = force_noise * abs(sustf.tst_suscept)**2

    return noise



import os
import copy
import tempfile

import scipy.io

from . import logger
from . import const
from . import suspension
from .struct import Struct

##################################################

MATLAB_ENGINE = None

class Matlab:
    def __init__(self, gwincpath=None):
        """Start a MATLAB engine for GWINC processing

         engine is provided, the GWINC path is added
        to it's path.

        """
        global MATLAB_ENGINE

        if MATLAB_ENGINE:
            return

        import matlab.engine

        if not gwincpath:
            gwincpath = os.path.expanduser(os.getenv('GWINCPATH', 'gwinc'))
        if not os.path.exists(os.path.join(gwincpath, 'gwinc.m')):
            raise IOError("Invalid MATLAB GWINC path: '{}'".format(gwincpath))

        logger.info("starting MATLAB engine...")
        MATLAB_ENGINE = matlab.engine.start_matlab()

        logger.info("gwinc path: {}".format(gwincpath))
        MATLAB_ENGINE.addpath(gwincpath)


    @property
    def eng(self):
        return MATLAB_ENGINE

    @property
    def workspace(self):
        return MATLAB_ENGINE.workspace

    def addpath(self, *args):
        return MATLAB_ENGINE.addpath(*args)

    def eval(self, *args, **kwargs):
        return MATLAB_ENGINE.eval(*args, **kwargs)


    def load_array(self, var, array):
        """Load numpy array into workspace as vector.

        `var` is name of workspace variable, and `array` is numpy
        ndarray.

        """
        # this stupidity because you can't just give the engine a np.ndarray
        MATLAB_ENGINE.workspace[var] = array.tolist()
        MATLAB_ENGINE.eval('{0} = cell2mat({0});'.format(var), nargout=0)


    def load_struct(self, var, struct):
        """Load pygwinc.Struct array into workspace as vector.

        `var` is name of workspace variable, and `struct` is
        pygwinc.Struct.

        """
        # similar stupidity prevents this (this time recarrays in the dict):
        #matlab.workspace['ifo'] = ifo.to_dict(array=True)
        with tempfile.NamedTemporaryFile(suffix='.mat') as f:
            scipy.io.savemat(f.name, struct.to_dict(array=True))
            MATLAB_ENGINE.eval("{} = load('{}');".format(var, f.name), nargout=0)


    def extract(self, *wvars):
        """Extract workspace variables from engine.

        Returns dict with wvars as keys.

        """
        assert len(wvars) > 0
        with tempfile.NamedTemporaryFile(suffix='.mat') as f:
            MATLAB_ENGINE.save(f.name, *wvars, nargout=0)
            data = scipy.io.loadmat(f.name, squeeze_me=True, struct_as_record=False)
        if len(wvars) == 1:
            return data[wvars[0]]
        else:
            return data

##################################################

NOISE_NAME_MAP = {
    'Quantum': 'Quantum Vacuum',
    'Newtonian': 'Newtonian Gravity',
    'CoatBrown': 'Coating Brownian',
    'CoatTO': 'Coating Thermo-Optic',
    'SubBrown': 'Substrate Brownian',
    'SubTE': 'Substrate Thermo-Elastic',
    'SuspThermal': 'Suspension Thermal',
    'ResGas': 'Excess Gas',
}


def ifo_matlab_transform(ifo):
    """Prep the ifo structure for use with MATLAB gwinc

    * add "Constants" sub-Struct
    * copy Temp to Constants
    * change Suspension.FiberType string into number

    """
    # add constants
    CONSTS = {k:v for k, v in const.__dict__ if not k.startswith('__')}
    ifo.Constants = Struct(CONSTS)

    # copy tempurature into Constants
    ifo.Constants.Temp = ifo.Infrastructure.Temp

    # translate FiberType string to int
    ifo.Suspension.FiberType = suspension.FIBER_TYPES.index(ifo.Suspension.FiberType)

    return ifo


def _rename_noises(d):
    nd = {}
    for k,v in d.items():
        try:
            nk = NOISE_NAME_MAP[k]
        except KeyError:
            nk = k
        if isinstance(v, dict):
            nd[nk] = _rename_noises(v)
        else:
            nd[nk] = v
    return nd


def gwinc_matlab(f, ifoin, plot=False):
    """Execute gwinc in MATLAB with the specified ifo model.

    This uses the python matlab.engine (see Matlab class) to calculate
    noises with MATLAB gwinc (gwinc directory specified by GWINCPATH
    environment variable).

    Returns `score` (dict), `noises` (dict), and `ifo` (Struct) as
    returned from MATLAB.

    If `plot` is True will cause MATLAB to produce it's own plot for
    the noise budget.

    """
    ifo = copy.deepcopy(ifoin)

    ifo_matlab_transform(ifo)

    matlab = Matlab()

    matlab.load_array('f', f)
    matlab.load_struct('ifo', ifo)

    if plot:
        plot_flag = '3'
    else:
        plot_flag = '0'

    cmd = "[score, noises, ifo] = gwinc(f, [], ifo, SourceModel, {});".format(plot_flag)
    matlab.eval(cmd, nargout=0)

    data = matlab.extract('score', 'noises', 'ifo')

    score = data['score']
    mnoises = Struct.from_matstruct(data['noises']).to_dict()
    ##### blow out 'MirrorThermal' sub-dict
    for n,d in mnoises['MirrorThermal'].items():
        if n == 'Total':
            continue
        mnoises[n] = d
    del mnoises['MirrorThermal']
    #####
    noises = _rename_noises(mnoises)
    ifo = Struct.from_matstruct(data['ifo'])

    return score, noises, ifo



"""
"""
import gwinc
from gwinc import load_budget


def test_load(pprint, tpath_join, fpath_join):
    pprint(gwinc.IFOS)
    for ifo in gwinc.IFOS:
        B = load_budget(ifo)
        trace = B.run()
        fig = trace.plot()
        fig.savefig(tpath_join('budget_{}.pdf'.format(ifo)))




from __future__ import division
import os
import sys
import logging
import importlib
import numpy as np

try:
    from ._version import version as __version__
except ModuleNotFoundError:
    try:
        import setuptools_scm
        __version__ = setuptools_scm.get_version(fallback_version='?.?.?')
    # FIXME: fallback_version is not available in the buster version
    # (3.2.0-1)
    except (ModuleNotFoundError, TypeError, LookupError):
        __version__ = '?.?.?'
from .ifo import IFOS
from .struct import Struct
from .plot import plot_trace
from .plot import plot_budget
from .plot import plot_noise


logger = logging.getLogger('gwinc')


DEFAULT_FREQ = '5:3000:6000'


class InvalidFrequencySpec(Exception):
    pass


def freq_from_spec(spec=None):
    """logarithmicly spaced frequency array, based on specification string

    Specification string should be of form 'START:[NPOINTS:]STOP'.  If
    `spec` is an array, then the array is returned as-is, and if it's
    None the DEFAULT_FREQ specification is used.

    """
    if isinstance(spec, np.ndarray):
        return spec
    elif spec is None:
        spec = DEFAULT_FREQ
    fspec = spec.split(':')
    try:
        if len(fspec) == 2:
            fspec = fspec[0], DEFAULT_FREQ.split(':')[1], fspec[1]
        return np.logspace(
            np.log10(float(fspec[0])),
            np.log10(float(fspec[2])),
            int(fspec[1]),
        )
    except (ValueError, IndexError):
        raise InvalidFrequencySpec(f'Improper frequency specification: {spec}')


def load_module(name_or_path):
    """Load module from name or path.

    Return loaded module and module path.

    """
    if os.path.exists(name_or_path):
        path = name_or_path.rstrip('/')
        modname = os.path.splitext(os.path.basename(path))[0]
        if os.path.isdir(path):
            path = os.path.join(path, '__init__.py')
        spec = importlib.util.spec_from_file_location(modname, path)
        mod = importlib.util.module_from_spec(spec)
        sys.modules[mod.__name__] = mod
        spec.loader.exec_module(mod)
    else:
        mod = importlib.import_module(name_or_path)
    try:
        path = mod.__path__[0]
    except AttributeError:
        path = mod.__file__
    return mod, path


def load_budget(name_or_path, freq=None, bname=None):
    """Load GWINC Budget

    Accepts either the name of a built-in canonical budget (see
    gwinc.IFOS), the path to a budget package (directory) or module
    (ending in .py), or the path to an IFO Struct definition file (see
    gwinc.Struct).  If an IFO Struct is specified, the base "aLIGO"
    budget definition will be used.

    If `bname` is specified the Budget class with that name will be
    loaded from the budget module.  Otherwise, the Budget class with
    the same name as the budget module will be loaded.

    If the budget is a package directory which includes an 'ifo.yaml'
    file the ifo Struct will be loaded from that file and assigned to
    the budget.ifo attribute.  If a Struct definition file is provided
    the base aLIGO budget definition will be assumed.

    Returns the instantiated Budget object.  If a frequency array or
    frequency specification string (see `freq_from_spec()`) is
    provided, the budget will be instantiated with the provided array.
    If a frequency array is not provided and the Budget class
    definition includes a `freq` attribute defining either an array or
    specification string, then that array will be used.  Otherwise a
    default array will be provided (see DEFAULT_FREQ).

    Any included ifo will be assigned as an attribute to the returned
    Budget object.

    """
    ifo = None

    if os.path.exists(name_or_path):
        path = name_or_path.rstrip('/')
        base, ext = os.path.splitext(os.path.basename(path))

        if ext in Struct.STRUCT_EXT:
            logger.info("loading struct {}...".format(path))
            ifo = Struct.from_file(path, _pass_inherit=True)

            inherit_ifo = ifo.get('+inherit', None)
            if inherit_ifo is not None:
                del ifo['+inherit']
                # make the inherited path relative to the loaded path
                # if it is a yml file
                if os.path.splitext(inherit_ifo)[1] in Struct.STRUCT_EXT:
                    base = os.path.split(path)[0]
                    inherit_ifo = os.path.join(base, inherit_ifo)

                inherit_budget = load_budget(inherit_ifo, freq=freq, bname=bname)
                pre_ifo = inherit_budget.ifo
                pre_ifo.update(ifo, overwrite_atoms=False)
                inherit_budget.update(ifo=pre_ifo)
                return inherit_budget
            else:
                modname = 'gwinc.ifo.aLIGO'
        else:
            bname = bname or base
            modname = path

    else:
        if name_or_path not in IFOS:
            raise RuntimeError("Unknown IFO '{}' (available IFOs: {}).".format(
                name_or_path,
                IFOS,
            ))
        bname = bname or name_or_path
        modname = 'gwinc.ifo.' + name_or_path

    logger.info(f"loading budget '{bname}' from {modname}...")
    mod, modpath = load_module(modname)
    Budget = getattr(mod, bname)
    if freq is None:
        freq = getattr(Budget, 'freq', None)
    freq = freq_from_spec(freq)
    ifopath = os.path.join(modpath, 'ifo.yaml')
    if not ifo and os.path.exists(ifopath):
        ifo = Struct.from_file(ifopath)
    return Budget(freq=freq, ifo=ifo)


def gwinc(freq, ifo, source=None, plot=False, PRfixed=True):
    """Calculate strain noise budget for a specified interferometer model.

    Argument `freq` is the frequency array for which the noises will
    be calculated, and `ifo` is the IFO model (see the `load_budget()`
    function).  The nominal 'aLIGO' budget structure will be used.

    If `source` structure provided, so evaluates the sensitivity of
    the detector to several potential gravitational wave
    sources.

    If `plot` is True a plot of the budget will be created.

    Returns tuple of (score, noises, ifo)

    """
    # assume generic aLIGO configuration
    # FIXME: how do we allow adding arbitrary addtional noise sources
    # from just ifo description, without having to specify full budget
    budget = load_budget('aLIGO', freq)
    traces = budget.run()
    plot_style = getattr(budget, 'plot_style', {})

    # construct matgwinc-compatible noises structure
    noises = {}
    for name, trace in traces.items():
        noises[name] = trace.psd
    noises['Total'] = traces.psd
    noises['Freq'] = traces.freq

    pbs = ifo.gwinc.pbs
    parm = ifo.gwinc.parm
    finesse = ifo.gwinc.finesse
    prfactor = ifo.gwinc.prfactor
    if ifo.Laser.Power * prfactor != pbs:
        logger.warning("Thermal lensing limits input power to {} W".format(pbs/prfactor))

    # report astrophysical scores if so desired
    score = None
    if source:
        logger.warning("Source score calculation currently not supported.  See `inspiral-range` package for similar functionality:")
        logger.warning("https://git.ligo.org/gwinc/inspiral-range")
        # score = int731(freq, noises['Total'], ifo, source)
        # score.Omega = intStoch(freq, noises['Total'], 0, ifo, source)

    # --------------------------------------------------------
    # output graphics

    if plot:
        logger.info('Laser Power:            %7.2f Watt' % ifo.Laser.Power)
        logger.info('SRM Detuning:           %7.2f degree' % (ifo.Optics.SRM.Tunephase*180/np.pi))
        logger.info('SRM transmission:       %9.4f' % ifo.Optics.SRM.Transmittance)
        logger.info('ITM transmission:       %9.4f' % ifo.Optics.ITM.Transmittance)
        logger.info('PRM transmission:       %9.4f' % ifo.Optics.PRM.Transmittance)
        logger.info('Finesse:                %7.2f' % finesse)
        logger.info('Power Recycling Gain:   %7.2f' % prfactor)
        logger.info('Arm Power:              %7.2f kW' % (parm/1000))
        logger.info('Power on BS:            %7.2f W' % pbs)

        # coating and substrate thermal load on the ITM
        PowAbsITM = (
            (pbs/2)
            * np.hstack([
                (finesse*2/np.pi) * ifo.Optics.ITM.CoatingAbsorption,
                (2 * ifo.Materials.MassThickness) * ifo.Optics.ITM.SubstrateAbsorption])
        )

        logger.info('Thermal load on ITM:    %8.3f W' % sum(PowAbsITM))
        logger.info('Thermal load on BS:     %8.3f W' % (ifo.Materials.MassThickness*ifo.Optics.SubstrateAbsorption*pbs))
        if (ifo.Laser.Power*prfactor != pbs):
            logger.info('Lensing limited input power: %7.2f W' % (pbs/prfactor))

        if score:
            logger.info('BNS Inspiral Range:     ' + str(score.effr0ns) + ' Mpc/ z = ' + str(score.zHorizonNS))
            logger.info('BBH Inspiral Range:     ' + str(score.effr0bh) + ' Mpc/ z = ' + str(score.zHorizonBH))
            logger.info('Stochastic Omega: %4.1g Universes' % score.Omega)

        traces.plot(**plot_style)

    return score, noises, ifo



import copy
import numpy as np
from numpy import pi, sin, exp, sqrt

from .. import logger
from .. import const
from ..struct import Struct
from .. import nb
from .. import noise
from .. import suspension


############################################################
# helper functions
############################################################

def mirror_struct(ifo, tm):
    """Create a "mirror" Struct for a LIGO core optic

    This is a copy of the ifo.Materials Struct, containing Substrate
    and Coating sub-Structs, as well as some basic geometrical
    properties of the optic.

    """
    # NOTE: we deepcopy this struct since we'll be modifying it (and
    # it would otherwise be stored by reference)
    mirror = copy.deepcopy(ifo.Materials)
    optic = ifo.Optics.get(tm)
    if 'CoatLayerOpticalThickness' in optic:
        mirror.Coating.dOpt = optic['CoatLayerOpticalThickness']
    else:
        T = optic.Transmittance
        dL = optic.CoatingThicknessLown
        dCap = optic.CoatingThicknessCap
        mirror.Coating.dOpt = noise.coatingthermal.getCoatDopt(mirror, T, dL, dCap=dCap)
    mirror.update(optic)
    mirror.MassVolume = pi * mirror.MassRadius**2 * mirror.MassThickness
    mirror.MirrorMass = mirror.MassVolume * mirror.Substrate.MassDensity
    return mirror


def arm_cavity(ifo):
    L = ifo.Infrastructure.Length

    g1 = 1 - L / ifo.Optics.Curvature.ITM
    g2 = 1 - L / ifo.Optics.Curvature.ETM
    gcav = sqrt(g1 * g2 * (1 - g1 * g2))
    gden = g1 - 2 * g1 * g2 + g2

    if (g1 * g2 * (1 - g1 * g2)) <= 0:
        raise Exception('Unstable arm cavity g-factors.  Change ifo.Optics.Curvature')
    elif gcav < 1e-3:
        logger.warning('Nearly unstable arm cavity g-factors.  Reconsider ifo.Optics.Curvature')

    ws = sqrt(L * ifo.Laser.Wavelength / pi)
    w1 = ws * sqrt(abs(g2) / gcav)
    w2 = ws * sqrt(abs(g1) / gcav)

    # waist size
    w0 = ws * sqrt(gcav / abs(gden))
    zr = pi * w0**2 / ifo.Laser.Wavelength
    z1 = L * g2 * (1 - g1) / gden
    z2 = L * g1 * (1 - g2) / gden

    # waist, input, output
    cavity = Struct()
    cavity.w0 = w0
    cavity.wBeam_ITM = w1
    cavity.wBeam_ETM = w2
    cavity.zr = zr
    cavity.zBeam_ITM = z1
    cavity.zBeam_ETM = z2
    return cavity


def ifo_power(ifo, PRfixed=True):
    """Compute power on beamsplitter, finesse, and power recycling factor.

    """
    pin = ifo.Laser.Power
    t1 = sqrt(ifo.Optics.ITM.Transmittance)
    r1 = sqrt(1 - ifo.Optics.ITM.Transmittance)
    r2 = sqrt(1 - ifo.Optics.ETM.Transmittance)
    t5 = sqrt(ifo.Optics.PRM.Transmittance)
    r5 = sqrt(1 - ifo.Optics.PRM.Transmittance)
    loss = ifo.Optics.Loss  # single TM loss
    bsloss = ifo.Optics.BSLoss
    acoat = ifo.Optics.ITM.CoatingAbsorption
    pcrit = ifo.Optics.pcrit

    # Finesse, effective number of bounces in cavity, power recycling factor
    finesse = 2*pi / (t1**2 + 2*loss)  # arm cavity finesse
    neff = 2 * finesse / pi

    # Arm cavity reflectivity with finite loss
    garm = t1 / (1 - r1*r2*sqrt(1-2*loss))  # amplitude gain wrt input field
    rarm = r1 - t1 * r2 * sqrt(1-2*loss) * garm

    if PRfixed:
        Tpr = ifo.Optics.PRM.Transmittance  # use given value
    else:
        Tpr = 1-(rarm*sqrt(1-bsloss))**2  # optimal recycling mirror transmission
        t5 = sqrt(Tpr)
        r5 = sqrt(1 - Tpr)
    prfactor = t5**2 / (1 + r5 * rarm * sqrt(1-bsloss))**2

    pbs = pin * prfactor  # BS power from input power
    parm = pbs * garm**2 / 2  # arm power from BS power

    thickness = ifo.Optics.ITM.get('Thickness', ifo.Materials.MassThickness)
    asub = 1.3 * 2 * thickness * ifo.Optics.SubstrateAbsorption
    pbsl = 2 * pcrit / (asub+acoat*neff)  # bs power limited by thermal lensing

    if pbs > pbsl:
        logger.warning('P_BS exceeds BS Thermal limit!')

    power = Struct()
    power.pbs = pbs
    power.parm = parm
    power.finesse = finesse
    power.gPhase = finesse * 2/np.pi
    power.prfactor = prfactor
    power.Tpr = Tpr
    return power

##################################################

def precomp_suspension(f, ifo):
    pc = Struct()
    pc.VHCoupling = Struct()
    if 'VHCoupling' in ifo.Suspension:
        pc.VHCoupling.theta = ifo.Suspension.VHCoupling.theta
    else:
        pc.VHCoupling.theta = ifo.Infrastructure.Length / const.R_earth
    hForce, vForce, hTable, vTable, tst_suscept = suspension.suspQuad(
        f, ifo.Suspension)
    pc.hForce = hForce
    pc.vForce = vForce
    pc.hTable = hTable
    pc.vTable = vTable
    pc.tst_suscept = tst_suscept
    return pc


@nb.precomp(sustf=precomp_suspension)
def precomp_quantum(f, ifo, sustf):
    pc = Struct()
    power = ifo_power(ifo)
    noise_dict = noise.quantum.shotrad(f, ifo, sustf, power)
    pc.ASvac = noise_dict['ASvac']
    pc.SEC = noise_dict['SEC']
    pc.Arm = noise_dict['arm']
    pc.Injection = noise_dict['injection']
    pc.PD = noise_dict['pd']

    # FC0 are the noises from the filter cavity losses and FC0_unsqzd_back
    # are noises from the unsqueezed vacuum injected at the back mirror
    # Right now there are at most one filter cavity in all the models;
    # if there were two, there would also be FC1 and FC1_unsqzd_back, etc.
    # keys = list(noise_dict.keys())
    fc_keys = [key for key in noise_dict.keys() if 'FC' in key]
    if fc_keys:
        pc.FC = np.zeros_like(pc.ASvac)
        for key in fc_keys:
            pc.FC += noise_dict[key]

    if 'phase' in noise_dict.keys():
        pc.Phase = noise_dict['phase']

    if 'ofc' in noise_dict.keys():
        pc.OFC = noise_dict['OFC']

    return pc


############################################################
# calibration
############################################################

def dhdl(f, armlen):
    """Strain to length conversion for noise power spetra

    This takes into account the GW wavelength and is only important
    when this is comparable to the detector arm length.

    From R. Schilling, CQG 14 (1997) 1513-1519, equation 5,
    with n = 1, nu = 0.05, ignoring overall phase and cos(nu)^2.
    A small value of nu is used instead of zero to avoid infinities.

    Returns the square of the dh/dL function, and the same divided by
    the arm length squared.

    """
    c = const.c
    nu_small = 15*pi/180
    omega_arm = pi * f * armlen / c
    omega_arm_f = (1 - sin(nu_small)) * pi * f * armlen / c
    omega_arm_b = (1 + sin(nu_small)) * pi * f * armlen / c
    sinc_sqr = 4 / abs(sin(omega_arm_f) * exp(-1j * omega_arm) / omega_arm_f
                       + sin(omega_arm_b) * exp(1j * omega_arm) / omega_arm_b)**2
    dhdl_sqr = sinc_sqr / armlen**2
    return dhdl_sqr, sinc_sqr


class Strain(nb.Calibration):
    """Calibrate displacement to strain
    """
    def calc(self):
        dhdl_sqr, sinc_sqr = dhdl(self.freq, self.ifo.Infrastructure.Length)
        return dhdl_sqr


class Force(nb.Calibration):
    """Calibrate displacement to force
    """
    def calc(self):
        mass = mirror_struct(self.ifo, 'ETM').MirrorMass
        return (mass * (2*pi*self.freq)**2)**2


############################################################
# noise sources
############################################################

#########################
# quantum
#########################

class QuantumVacuum(nb.Noise):
    """Quantum Vacuum

    """
    style = dict(
        label='Quantum Vacuum',
        color='#ad03de',
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        total = np.zeros_like(quantum.ASvac)
        for nn in quantum.values():
            total += nn
        return total


class QuantumVacuumAS(nb.Noise):
    """Quantum vacuum from the AS port

    """
    style = dict(
        label='AS Port Vacuum',
        color='xkcd:emerald green'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.ASvac


class QuantumVacuumArm(nb.Noise):
    """Quantum vacuum due to arm cavity loss

    """
    style = dict(
        label='Arm Loss',
        color='xkcd:orange brown'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.Arm


class QuantumVacuumSEC(nb.Noise):
    """Quantum vacuum due to SEC loss

    """
    style = dict(
        label='SEC Loss',
        color='xkcd:cerulean'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.SEC


class QuantumVacuumFilterCavity(nb.Noise):
    """Quantum vacuum due to filter cavity loss

    """
    style = dict(
        label='Filter Cavity Loss',
        color='xkcd:goldenrod'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.FC


class QuantumVacuumInjection(nb.Noise):
    """Quantum vacuum due to injection loss

    """
    style = dict(
        label='Injection Loss',
        color='xkcd:fuchsia'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.Injection


class QuantumVacuumReadout(nb.Noise):
    """Quantum vacuum due to readout loss

    """
    style = dict(
        label='Readout Loss',
        color='xkcd:mahogany'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.PD


class QuantumVacuumQuadraturePhase(nb.Noise):
    """Quantum vacuum noise due to quadrature phase noise
    """
    style = dict(
        label='Quadrature Phase',
        color='xkcd:slate'
    )

    @nb.precomp(quantum=precomp_quantum)
    def calc(self, quantum):
        return quantum.Phase


class StandardQuantumLimit(nb.Noise):
    """Standard Quantum Limit

    """
    style = dict(
        label="Standard Quantum Limit",
        color="#000000",
        linestyle=":",
    )

    def calc(self):
        ETM = mirror_struct(self.ifo, 'ETM')
        return 8 * const.hbar / (ETM.MirrorMass * (2 * np.pi * self.freq) ** 2)


#########################
# seismic
#########################

class SeismicHorizontal(nb.Noise):
    """Horizontal seismic noise

    """
    style = dict(
        label='Horizontal',
        color='xkcd:muted blue',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        nt, nr = noise.seismic.platform_motion(self.freq, self.ifo)
        n = noise.seismic.seismic_suspension_filtered(sustf, nt, 'horiz')
        return n * 4


class SeismicVertical(nb.Noise):
    """Vertical seismic noise

    """
    style = dict(
        label='Vertical',
        color='xkcd:brick red',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        nt, nr = noise.seismic.platform_motion(self.freq, self.ifo)
        n = noise.seismic.seismic_suspension_filtered(sustf, nt, 'vert')
        return n * 4


class Seismic(nb.Budget):
    """Seismic

    """
    style = dict(
        label='Seismic',
        color='#855700',
    )

    noises = [
        SeismicHorizontal,
        SeismicVertical,
    ]


#########################
# Newtonian
#########################

class Newtonian(nb.Noise):
    """Newtonian Gravity

    """
    style = dict(
        label='Newtonian Gravity',
        color='#15b01a',
    )

    def calc(self):
        n = noise.newtonian.gravg(self.freq, self.ifo.Seismic)
        return n * 4


class NewtonianRayleigh(nb.Noise):
    """Newtonian Gravity, Rayleigh waves

    """
    style = dict(
        label='Newtonian (Rayleigh waves)',
        color='#1b2431',
    )

    def calc(self):
        n = noise.newtonian.gravg_rayleigh(self.freq, self.ifo.Seismic)
        return n * 2


class NewtonianBody(nb.Noise):
    """Newtonian Gravity, body waves

    """
    style = dict(
        label='Newtonian (body waves)',
        color='#85a3b2',
    )

    def calc(self):
        np = noise.newtonian.gravg_pwave(self.freq, self.ifo.Seismic)
        ns = noise.newtonian.gravg_swave(self.freq, self.ifo.Seismic)
        return (np + ns) * 4


class NewtonianInfrasound(nb.Noise):
    """Newtonian Gravity, infrasound

    """
    style = dict(
        label='Newtonian (infrasound)',
        color='#ffa62b',
    )

    def calc(self):
        n = noise.newtonian.atmois(self.freq, self.ifo.Atmospheric, self.ifo.Seismic)
        return n * 2


#########################
# suspension thermal
#########################

class SuspensionThermalHorizTop(nb.Noise):
    """Horizontal suspension thermal around the top mass

    """
    style = dict(
        label='Horiz. Top',
        color='xkcd:orangeish',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 0, 'horiz')
        return abs(n) * 4


class SuspensionThermalHorizAPM(nb.Noise):
    """Horizontal suspension thermal around the upper intermediate mass

    """
    style = dict(
        label='Horiz. APM',
        color='xkcd:mustard',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 1, 'horiz')
        return abs(n) * 4


class SuspensionThermalHorizPUM(nb.Noise):
    """Horizontal suspension thermal around the penultimate mass

    """
    style = dict(
        label='Horiz. PUM',
        color='xkcd:turquoise',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 2, 'horiz')
        return abs(n) * 4


class SuspensionThermalHorizTM(nb.Noise):
    """Horizontal suspension thermal around the test

    """
    style = dict(
        label='Horiz. Test mass',
        color='xkcd:bright purple',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 3, 'horiz')
        return abs(n) * 4


class SuspensionThermalVertTop(nb.Noise):
    """Vertical suspension thermal around the top mass

    """
    style = dict(
        label='Vert. Top',
        color='xkcd:orangeish',
        linestyle='--',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 0, 'vert')
        return abs(n) * 4


class SuspensionThermalVertAPM(nb.Noise):
    """Vertical suspension thermal around the upper intermediate mass

    """
    style = dict(
        label='Vert. APM',
        color='xkcd:mustard',
        linestyle='--',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 1, 'vert')
        return abs(n) * 4


class SuspensionThermalVertPUM(nb.Noise):
    """Vertical suspension thermal around the penultimate mass

    """
    style = dict(
        label='Vert. PUM',
        color='xkcd:turquoise',
        linestyle='--',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 2, 'vert')
        return abs(n) * 4


class SuspensionThermalVertTM(nb.Noise):
    """Vertical suspension thermal around the test

    """
    style = dict(
        label='Vert. Test mass',
        color='xkcd:bright purple',
        linestyle='--',
        alpha=0.7,
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        n = noise.suspensionthermal.susptherm_stage(
            self.freq, self.ifo.Suspension, sustf, 3, 'vert')
        return abs(n) * 4


class SuspensionThermal(nb.Budget):
    """Suspension Thermal

    """

    name = 'SuspensionThermal'

    style = dict(
        label='Suspension Thermal',
        color='#0d75f8',
    )

    noises = [
        SuspensionThermalHorizTop,
        SuspensionThermalHorizAPM,
        SuspensionThermalHorizPUM,
        SuspensionThermalHorizTM,
        SuspensionThermalVertTop,
        SuspensionThermalVertAPM,
        SuspensionThermalVertPUM,
    ]


#########################
# coating thermal
#########################

class CoatingBrownian(nb.Noise):
    """Coating Brownian

    """
    style = dict(
        label='Coating Brownian',
        color='#fe0002',
    )

    def calc(self):
        ITM = mirror_struct(self.ifo, 'ITM')
        ETM = mirror_struct(self.ifo, 'ETM')
        cavity = arm_cavity(self.ifo)
        wavelength = self.ifo.Laser.Wavelength
        nITM = noise.coatingthermal.coating_brownian(
            self.freq, ITM, wavelength, cavity.wBeam_ITM
        )
        nETM = noise.coatingthermal.coating_brownian(
            self.freq, ETM, wavelength, cavity.wBeam_ETM
        )
        return (nITM + nETM) * 2


class CoatingThermoOptic(nb.Noise):
    """Coating Thermo-Optic

    """
    style = dict(
        label='Coating Thermo-Optic',
        color='#02ccfe',
        linestyle='--',
    )

    def calc(self):
        wavelength = self.ifo.Laser.Wavelength
        materials = self.ifo.Materials
        ITM = mirror_struct(self.ifo, 'ITM')
        ETM = mirror_struct(self.ifo, 'ETM')
        cavity = arm_cavity(self.ifo)
        nITM, junk1, junk2, junk3 = noise.coatingthermal.coating_thermooptic(
            self.freq, ITM, wavelength, cavity.wBeam_ITM,
        )
        nETM, junk1, junk2, junk3 = noise.coatingthermal.coating_thermooptic(
            self.freq, ETM, wavelength, cavity.wBeam_ETM,
        )
        return (nITM + nETM) * 2


#########################
# substrate thermal
#########################

class ITMThermoRefractive(nb.Noise):
    """ITM Thermo-Refractive

    """
    style = dict(
        label='ITM Thermo-Refractive',
        color='#448ee4',
        linestyle='--',
    )

    def calc(self):
        power = ifo_power(self.ifo)
        gPhase = power.finesse * 2/np.pi
        cavity = arm_cavity(self.ifo)
        n = noise.substratethermal.substrate_thermorefractive(
            self.freq, self.ifo.Materials, cavity.wBeam_ITM)
        return n * 2 / gPhase**2


class SubstrateBrownian(nb.Noise):
    """Substrate Brownian

    """
    style = dict(
        label='Substrate Brownian',
        color='#fb7d07',
        linestyle='--',
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        nITM = noise.substratethermal.substrate_brownian(
            self.freq, self.ifo.Materials, cavity.wBeam_ITM)
        nETM = noise.substratethermal.substrate_brownian(
            self.freq, self.ifo.Materials, cavity.wBeam_ETM)
        return (nITM + nETM) * 2


class SubstrateThermoElastic(nb.Noise):
    """Substrate Thermo-Elastic

    """
    style = dict(
        label='Substrate Thermo-Elastic',
        color='#f5bf03',
        linestyle='--',
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        nITM = noise.substratethermal.substrate_thermoelastic(
            self.freq, self.ifo.Materials, cavity.wBeam_ITM)
        nETM = noise.substratethermal.substrate_thermoelastic(
            self.freq, self.ifo.Materials, cavity.wBeam_ETM)
        return (nITM + nETM) * 2


#########################
# residual gas
#########################

# FIXME HACK: it's unclear if a phase noise in the arms like
# the excess gas noise should get the same dhdL strain
# calibration as the other displacement noises.  However, we
# would like to use the one Strain calibration for all noises,
# so we need to divide by the sinc_sqr here to undo the
# application of the dhdl in the Strain calibration.  But this
# is ultimately a superfluous extra calculation with the only
# to provide some simplication in the Budget definition, so
# should be re-evaluated at some point.


def calc_total_residual_gas_damping(f, ifo, species, sustf):
    """Total residual gas damping from all four test masses

    :f: frequency array in Hz
    :ifo: gwinc IFO structure
    :species: molecular species structure
    :sustf: suspension transfer function structure

    :returns: displacement noise
    """
    squeezed_film = ifo.Infrastructure.ResidualGas.get('SqueezedFilm', Struct())

    if squeezed_film is None:
        raise ValueError('Must specify either excess damping or a gap')

    # Calculate squeezed film for ETM and ITM seperately if either is given
    # explicitly. If only one is given, it is not computed for the other one.
    if ('ETM' in squeezed_film) or ('ITM' in squeezed_film):
        squeezed_film_ETM = squeezed_film.get('ETM', Struct())
        squeezed_film_ITM = squeezed_film.get('ITM', Struct())
        n_ETM = noise.residualgas.residual_gas_damping_test_mass(
            f, ifo, species, sustf, squeezed_film_ETM)
        n_ITM = noise.residualgas.residual_gas_damping_test_mass(
            f, ifo, species, sustf, squeezed_film_ITM)
        n = 2 * (n_ETM + n_ITM)

    # Otherwise the same calculation is used for both.
    else:
        n = 4 * noise.residualgas.residual_gas_damping_test_mass(
            f, ifo, species, sustf, squeezed_film)

    return n


class ExcessGasScatteringH2(nb.Noise):
    """Excess gas scattering for H2

    """
    style = dict(
        label='H$_2$ scattering',
        color='xkcd:red orange'
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        species = self.ifo.Infrastructure.ResidualGas.H2
        n = noise.residualgas.residual_gas_scattering_arm(
            self.freq, self.ifo, cavity, species)
        dhdl_sqr, sinc_sqr = dhdl(self.freq, self.ifo.Infrastructure.Length)
        return n * 2 / sinc_sqr


class ExcessGasScatteringN2(nb.Noise):
    """Excess gas scattering for N2

    """
    style = dict(
        label='N$_2$ scattering',
        color='xkcd:emerald'
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        species = self.ifo.Infrastructure.ResidualGas.N2
        n = noise.residualgas.residual_gas_scattering_arm(
            self.freq, self.ifo, cavity, species)
        dhdl_sqr, sinc_sqr = dhdl(self.freq, self.ifo.Infrastructure.Length)
        return n * 2 / sinc_sqr


class ExcessGasScatteringH2O(nb.Noise):
    """Excess gas scattering for H2O

    """
    style = dict(
        label='H$_2$O scattering',
        color='xkcd:water blue'
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        species = self.ifo.Infrastructure.ResidualGas.H2O
        n = noise.residualgas.residual_gas_scattering_arm(
            self.freq, self.ifo, cavity, species)
        dhdl_sqr, sinc_sqr = dhdl(self.freq, self.ifo.Infrastructure.Length)
        return n * 2 / sinc_sqr


class ExcessGasScatteringO2(nb.Noise):
    """Excess gas scattering for O2

    """
    style = dict(
        label='O$_2$ scattering',
        color='xkcd:grey'
    )

    def calc(self):
        cavity = arm_cavity(self.ifo)
        species = self.ifo.Infrastructure.ResidualGas.O2
        n = noise.residualgas.residual_gas_scattering_arm(
            self.freq, self.ifo, cavity, species)
        dhdl_sqr, sinc_sqr = dhdl(self.freq, self.ifo.Infrastructure.Length)
        return n * 2 / sinc_sqr


class ExcessGasDampingH2(nb.Noise):
    """Residual gas damping for H2

    """
    style = dict(
        label='H$_2$ damping',
        color='xkcd:red orange',
        linestyle='--',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        species = self.ifo.Infrastructure.ResidualGas.H2
        return calc_total_residual_gas_damping(self.freq, self.ifo, species, sustf)


class ExcessGasDampingN2(nb.Noise):
    """Excess gas damping for N2

    """
    style = dict(
        label='N$_2$ damping',
        color='xkcd:emerald',
        linestyle='--',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        species = self.ifo.Infrastructure.ResidualGas.N2
        return calc_total_residual_gas_damping(self.freq, self.ifo, species, sustf)


class ExcessGasDampingH2O(nb.Noise):
    """Excess gas damping for H2O

    """
    style = dict(
        label='H$_2$O damping',
        color='xkcd:water blue',
        linestyle='--',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        species = self.ifo.Infrastructure.ResidualGas.H2O
        return calc_total_residual_gas_damping(self.freq, self.ifo, species, sustf)


class ExcessGasDampingO2(nb.Noise):
    """Excess gas damping for O2

    """
    style = dict(
        label='O$_2$ damping',
        color='xkcd:grey',
        linestyle='--',
    )

    @nb.precomp(sustf=precomp_suspension)
    def calc(self, sustf):
        species = self.ifo.Infrastructure.ResidualGas.O2
        return calc_total_residual_gas_damping(self.freq, self.ifo, species, sustf)


class ExcessGas(nb.Budget):
    """Excess Gas

    """
    style = dict(
        label='Residual Gas',
        color='#add00d',
        linestyle='-',
    )

    noises = [
        ExcessGasScatteringH2,
        ExcessGasScatteringN2,
        ExcessGasScatteringH2O,
        ExcessGasScatteringO2,
        ExcessGasDampingH2,
        ExcessGasDampingN2,
        ExcessGasDampingH2O,
        ExcessGasDampingO2,
    ]



"""
"""
import numpy as np
from os import path
from gwinc import Struct
import pylab as pyl


def test_load(pprint, tpath_join, fpath_join):
    fpath = fpath_join('test_load.yml')
    yml = Struct.from_file(fpath)
    pprint("full yaml")
    pprint(yml)
    pprint("individual tests")
    pprint(yml.lists.test_nonelist)
    assert(yml.lists.test_nonelist == [None, None])
    pprint(yml.dicts.test_nonedict)
    S_cmp = Struct({'A': None, 'B': None})
    pprint(yml.dicts.test_nonedict.diff(S_cmp))
    assert(yml.dicts.test_nonedict == S_cmp)


def test_struct_load_Sequences():
    # test of the ordering of value types

    # https://git.ligo.org/gwinc/pygwinc/-/merge_requests/127#note_458256
    s = Struct({'a': 'b'})
    s.update({'a': 'c'})

    return





from numpy import pi, sqrt

from . import const
from .ifo.noises import ifo_power


def sql(ifo):
    """Computer standard quantum limit (SQL) for IFO"""
    c = const.c
    power = ifo_power(ifo)
    w0 = 2 * pi * c / ifo.Laser.Wavelength
    m = ifo.Suspension.Stage[0].Mass
    Titm = ifo.Optics.ITM.Transmittance
    Tsrm = ifo.Optics.SRM.Transmittance
    tSR = sqrt(Tsrm)
    rSR = sqrt(1 - Tsrm)
    fSQL = (1/(2*pi))*(8/c)*sqrt((power.parm*w0)/(m*Titm))*(tSR/(1+rSR))
    return fSQL


def computeFCParams(ifo):
    """Compute ideal filter cavity Tin, detuning [Hz] and bandwidth [Hz]

    """
    # FC parameters
    fcParams = ifo.Squeezer.FilterCavity
    c = const.c
    fsrFC = c / (2 * fcParams.L)
    lossFC = fcParams.Lrt + fcParams.Te

    fSQL = sql(ifo)

    # detuning and cavity bandwidth (D&D paper P1400018 and/or PRD)
    eps = 4 / (2 + sqrt(2 + 2 * sqrt(1 + (4 * pi * fSQL / (fsrFC * lossFC))**4)))
    s1eps = sqrt(1 - eps)

    # cavity bandwidth [Hz]
    gammaFC = fSQL / sqrt(s1eps + s1eps**3)
    # cavity detuning [Hz]
    detuneFC = s1eps * gammaFC

    # input mirror transmission
    TinFC = 4 * pi * gammaFC / fsrFC - lossFC
    if TinFC < lossFC:
        raise RuntimeError(
            'IFC: Losses are too high! {:0.1f} ppm max.'.format(1e6 * gammaFC / fsrFC))

    # Add to fcParams structure
    fcParams.Ti = TinFC
    fcParams.fdetune = -detuneFC
    fcParams.gammaFC = gammaFC
    fcParams.fsrFC = fsrFC

    return fcParams



import os


IFOS = [
    'aLIGO',
    'Aplus',
    'Voyager',
    'CE1',
    'CE2silica',
    'CE2silicon',
]


PLOT_STYLE = dict(
    ylabel=u"Strain [1/\u221AHz]",
)


def lpath(file0, file1):
    """Return path of file1 when expressed relative to file0.

    For instance, if file0 is "/path/to/file0" and file1 is
    "../for/file1" then what is returned is "/path/for/file1".

    This is useful for resolving paths within packages with e.g.:

      rpath = lpath(__file__, '../resource.yaml')

    """
    return os.path.abspath(os.path.join(os.path.dirname(file0), file1))



import os
import sys
import shutil
import signal
import logging
import tempfile
import argparse
import subprocess
import matplotlib.pyplot as plt
from collections import OrderedDict
from PyPDF2 import PdfFileReader, PdfFileWriter

from .. import IFOS, load_budget
from ..io import load_hdf5

logging.basicConfig(
    format='%(message)s',
    level=os.getenv('LOG_LEVEL', 'INFO').upper())

try:
    import inspiral_range
except ImportError:
    logging.warning("inspiral_range package not found, range will not be calculated")
    inspiral_range = None


TOLERANCE = 1e-6
CACHE_LIMIT = 5


def test_path(*args):
    """Return path to package file."""
    return os.path.join(os.path.dirname(__file__), *args)


def git_find_upstream_name():
    try:
        remotes = subprocess.run(
            ['git', 'remote', '-v'],
            capture_output=True, universal_newlines=True,
            check=True,
        ).stdout
    except subprocess.CalledProcessError as e:
        logging.error(e.stderr.split('\n')[0])
    for remote in remotes.strip().split('\n'):
        name, url, fp = remote.split()
        if 'gwinc/pygwinc.git' in url:
            return name


def git_rev_resolve_hash(git_rev):
    """Resolve a git revision into its hash string."""
    try:
        return subprocess.run(
            ['git', 'show', '-s', '--format=format:%H', git_rev],
            capture_output=True, universal_newlines=True,
            check=True,
        ).stdout
    except subprocess.CalledProcessError as e:
        logging.error(e.stderr.split('\n')[0])


def prune_cache_dir():
    """Prune all but the N most recently accessed caches.

    """
    cache_dir = test_path('cache')
    if not os.path.exists(cache_dir):
        return
    expired_paths = sorted(
        [os.path.join(cache_dir, path) for path in os.listdir(cache_dir)],
        key=lambda path: os.stat(path).st_atime, reverse=True,
    )[CACHE_LIMIT:]
    if not expired_paths:
        return
    for path in expired_paths:
        logging.info("pruning old cache: {}".format(path))
        shutil.rmtree(path)


def gen_cache(git_hash, path):
    """generate cache for specified git hash at the specified path

    The included shell script is used to extract the gwinc code from
    the appropriate git commit, and invoke a new python instance to
    generate the noise curves.

    """
    logging.info("creating new cache for hash {}...".format(git_hash))
    subprocess.run(
        [test_path('gen_cache.sh'), git_hash, path, sys.executable],
        check=True,
    )


def load_cache(path):
    """load a cache from the specified path

    returns a "cache" dictionary with 'git_hash' and 'ifos' keys.

    """
    logging.info("loading cache {}...".format(path))
    cache = {}
    git_hash_path = os.path.join(path, 'git_hash')
    if os.path.exists(git_hash_path):
        with open(git_hash_path) as f:
            git_hash = f.read().strip()
    else:
        git_hash = None
    logging.debug("cache hash: {}".format(git_hash))
    cache['git_hash'] = git_hash
    cache['ifos'] = {}
    for f in sorted(os.listdir(path)):
        name, ext = os.path.splitext(f)
        if ext != '.h5':
            continue
        cache['ifos'][name] = os.path.join(path, f)
    return cache


def zip_noises(traceA, traceB, skip):
    """zip matching noises from traces"""
    B_dict = dict(traceB.walk())
    for name, tA in traceA.walk():
        if skip and name in skip:
            logging.warning("SKIPPING TEST: '{}'".format(name))
            continue
        try:
            tB = B_dict[name]
        except (KeyError, TypeError):
            logging.warning("MISSING B TRACE: '{}'".format(name))
            continue
        yield name, tA, tB


def compare_traces(traceA, traceB, tolerance=TOLERANCE, skip=None):
    """Compare two gwinc traces

    Noises listed in `skip` will not be compared.

    Returns a dictionary of noises that differ fractionally (on a
    point-by-point basis) by more than `tolerance` between the traces
    in `traceA` and `traceB`.

    """
    #name_width = max([len(n[0][-1]) for n in walk_traces(tracesA)])
    name_width = 15
    diffs = OrderedDict()
    for name, tA, tB in zip_noises(traceA, traceB, skip):
        logging.debug("comparing {}...".format(name))
        ampA = tA.asd
        ampB = tB.asd
        diff = ampB - ampA
        frac = abs(diff / ampA)
        if max(frac) < tolerance:
            continue
        logging.warning("EXCESSIVE DIFFERENCE: {:{w}} {:6.1f} ppm".format(
            name, max(frac)*1e6, w=name_width))
        diffs[name] = (ampA, ampB, frac)
    return diffs


def plot_diffs(freq, diffs, styleA, styleB):
    spec = (len(diffs)+1, 2)
    sharex = None
    for i, nname in enumerate(diffs):
        ampA, ampB, frac = diffs[nname]

        axl = plt.subplot2grid(spec, (i, 0), sharex=None)
        axl.loglog(freq, ampA, **styleA)
        axl.loglog(freq, ampB, **styleB)
        axl.grid()
        axl.legend(loc='upper right')
        axl.set_ylabel(nname)
        if i == 0:
            axl.set_title("noise value")

        if i == 0:
            sharex = axl
        axr = plt.subplot2grid(spec, (i, 1), sharex=sharex)
        axr.loglog(freq, frac)
        axr.grid()
        axr.axhline(y=max(frac), color='r', linestyle='--')
        axr.text(max(freq)+4000, max(frac), '{:.1f} ppm'.format(max(frac)*1e6),
                 horizontalalignment='left', verticalalignment='center',
                 color='red')
        if i == 0:
            axr.set_title("fractional difference")

    axl.set_xlabel("frequency [Hz]")
    axr.set_xlabel("frequency [Hz]")
    plt.subplots_adjust(top=0.8, right=0.85, wspace=0.3)

##################################################

def main():
    parser = argparse.ArgumentParser(
        description="""GWINC noise validation

This command calculates the canonical noise budgets with the current
code and compares them against those calculated with code from a
specified git revision.  You must be running from a git checkout of
the source for this to work.  The command will fail if it detects any
noise differences.  Plots or a PDF report of differences can be
generated with the '--plot' or '--report' commands respectively.

By default it will attempt to determine the git reference for upstream
master for your current configuration (usually 'origin/master' or
'upstream/master').  You may specify an arbitrary git revision with
the --git-rev command.  For example, to compare against another
remote/branch use:

$ python3 -m gwinc.test --git-rev remote/dev-branch

or if you have uncommitted changes compare against the current head
with:

$ python3 -m gwinc.test -g HEAD

See gitrevisions(7) for various ways to refer to git revisions.

A cache of traces from reference git revisions will be stored in
gwinc/test/cache/<SHA1>.  Old caches are automatically pruned.""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '--tolerance', '-t',  type=float, default=TOLERANCE,
        help="fractional tolerance of comparison [{}]".format(TOLERANCE))
    parser.add_argument(
        '--skip', '-k', metavar='NOISE', action='append',
        help="traces to skip in comparison (multiple may be specified)")
    rgroup = parser.add_mutually_exclusive_group()
    rgroup.add_argument(
        '--git-rev', '-g', metavar='REV',
        help="specify specific git revision to compare against")
    rgroup.add_argument(
        '--head', '-gh', action='store_const', dest='git_rev', const='HEAD',
        help="shortcut for '--git-rev HEAD'")
    ogroup = parser.add_mutually_exclusive_group()
    ogroup.add_argument(
        '--plot', '-p', action='store_true',
        help="show interactive plot differences")
    ogroup.add_argument(
        '--report', '-r', metavar='REPORT.pdf',
        help="create PDF report of test results (only created if differences found)")
    parser.add_argument(
        'ifo', metavar='IFO', nargs='*',
        help="specific ifos to test (default all)")
    args = parser.parse_args()

    # get the reference hash
    if args.git_rev:
        git_rev = args.git_rev
    else:
        remote = git_find_upstream_name()
        if not remote:
            sys.exit("Could not resolve upstream remote name")
        git_rev = '{}/master'.format(remote)
    logging.info("git  rev: {}".format(git_rev))
    git_hash = git_rev_resolve_hash(git_rev)
    if not git_hash:
        sys.exit("Could not resolve reference, could not run test.")
    logging.info("git hash: {}".format(git_hash))

    # load the cache
    cache_path = test_path('cache', git_hash)
    if not os.path.exists(cache_path):
        prune_cache_dir()
        gen_cache(git_hash, cache_path)
    cache = load_cache(cache_path)

    if args.report:
        base, ext = os.path.splitext(args.report)
        if ext != '.pdf':
            parser.error("Test reports only support PDF format.")
        outdir = tempfile.TemporaryDirectory()

    if args.ifo:
        ifos = args.ifo
    else:
        ifos = IFOS

    style_ref = dict(label='reference', linestyle='-')
    style_cur = dict(label='current', linestyle='--')

    fail = False

    # compare
    for name in ifos:
        logging.info("{} tests...".format(name))

        try:
            path = cache['ifos'][name]
        except KeyError:
            logging.warning("IFO {} not found in cache")
            fail |= True
            continue

        traces_ref = load_hdf5(path)
        traces_ref.name = 'Total'
        freq = traces_ref.freq
        budget = load_budget(name, freq)
        traces_cur = budget.run()
        traces_cur.name = 'Total'

        # FIXME: add this back
        if False: #inspiral_range:
            total_ref = traces_ref.psd
            total_cur = traces_cur.psd
            range_func = inspiral_range.range
            H = inspiral_range.waveform.CBCWaveform(freq)
            fom_ref = range_func(freq, total_ref, H=H)
            traces_ref['int73'] = inspiral_range.int73(freq, total_ref)[1], None
            fom_cur = range_func(freq, total_cur, H=H)
            traces_cur['int73'] = inspiral_range.int73(freq, total_cur)[1], None
            fom_summary = """
inspiral {func} {m1}/{m2} Msol:
{label_ref}: {fom_ref:.2f} Mpc
{label_cur}: {fom_cur:.2f} Mpc
""".format(
                func=range_func.__name__,
                m1=H.params['m1'],
                m2=H.params['m2'],
                label_ref=style_ref['label'],
                fom_ref=fom_ref,
                label_cur=style_cur['label'],
                fom_cur=fom_cur,
            )
        else:
            fom_summary = ''

        diffs = compare_traces(traces_ref, traces_cur, args.tolerance, args.skip)

        if not diffs:
            logging.info("{} tests pass.".format(name))
            continue

        logging.warning("{} tests FAIL".format(name))
        fail |= True
        if args.plot or args.report:
            plot_diffs(freq, diffs, style_ref, style_cur)
            plt.suptitle('''{} {}/{} noise comparison
(noises that differ by more than {} ppm)
reference git hash: {}
{}'''.format(name, style_ref['label'], style_cur['label'],
             args.tolerance*1e6, cache['git_hash'], fom_summary))
            if args.report:
                pwidth = 10
                pheight = (len(diffs) * 5) + 2
                plt.gcf().set_size_inches(pwidth, pheight)
                plt.savefig(os.path.join(outdir.name, name+'.pdf'))
            else:
                plt.show()

    if not fail:
        logging.info("all tests pass.")
        return 0

    if args.report:
        logging.info("generating report {}...".format(args.report))
        pdf_writer = PdfFileWriter()
        for name in ifos:
            path = os.path.join(outdir.name, '{}.pdf'.format(name))
            if not os.path.exists(path):
                continue
            pdf_reader = PdfFileReader(path)
            for page in range(pdf_reader.getNumPages()):
                pdf_writer.addPage(pdf_reader.getPage(page))
        with open(args.report, 'wb') as f:
            pdf_writer.write(f)
        outdir.cleanup()

    logging.info("TESTS FAILED.")
    return 1

##################################################

if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    sys.exit(main())



import numpy as np

from .plot import plot_trace


class BudgetTrace:
    """Budget trace data calculated from a Noise or Budget


    """
    def __init__(self, name=None, style=None, freq=None, psd=None, budget=None):
        self.name = name
        self.style = style or {}
        self._freq = freq
        self._psd = psd
        self.budget = budget or []

    def __repr__(self):
        if self.budget:
            bs = ' [{}]'.format(', '.join([str(b.name) for b in self]))
        else:
            bs = ''
        return '<{} {}{}>'.format(
            self.__class__.__name__,
            self.name,
            bs,
        )

    @property
    def freq(self):
        """trace frequency array, in Hz"""
        return self._freq

    @property
    def psd(self):
        """trace power spectral density array"""
        return self._psd

    @property
    def asd(self):
        """trace amplitude spectral density array"""
        return np.sqrt(self._psd)

    def len(self):
        """Length of data"""
        return len(self._freq)

    def __iter__(self):
        """iterator of budget traces"""
        return iter(self.budget)

    @property
    def _bdict(self):
        bdict = {trace.name: trace for trace in self.budget}
        return bdict

    def __getattr__(self, name):
        try:
            return self._bdict[name]
        except KeyError:
            raise AttributeError

    def __getitem__(self, name):
        """get budget trace by name

        """
        try:
            name, rest = name.split('.', 1)
            return self._bdict[name][rest]
        except ValueError:
            return self._bdict[name]

    def get(self, key, default=None):
        """get a (possibly nested) Trace item.

        """
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        """iterator of budget (name, trace) tuples"""
        return self._bdict.items()

    def keys(self):
        """iterator of budget trace names"""
        return self._bdict.keys()

    def values(self):
        """iterator of budget traces"""
        return self._bdict.values()

    def walk(self):
        """walk recursively through all traces"""
        yield self.name, self
        for trace in self:
            for t in trace.walk():
                yield t


    def plot(self, ax=None, **kwargs):
        """Plot the trace budget

        If an axis handle `ax` is provided it will be used for the
        plot.  All remaining keyword arguments are assumed to define
        various matplotlib plot style attributes.

        Returns the figure handle.

        """
        return plot_trace(self, ax=ax, **kwargs)



"""
Requires pytest to import
"""
import os
from os import path
import os
from shutil import rmtree
import contextlib

import pytest

_options_added = False

def pytest_addoption(parser):
    global _options_added
    #this check fixes issues if this gets run multiple times from sub conftest.py's
    if _options_added:
        return
    else:
        _options_added = True

    parser.addoption(
        "--plot",
        action="store_true",
        dest = 'plot',
        help = "Have tests update plots (it is slow)",
    )

    parser.addoption(
        "--do-stresstest",
        action = "store_true",
        help   = "Run slow repeated stress tests"
    )

    parser.addoption(
        "--no-preclear",
        action="store_true",
        default=False,
        dest='no_preclear',
        help="Do not preclear tpaths",
    )


@pytest.fixture
def plot(request):
    return request.config.getvalue('--plot')
    return request.config.option.plot


@pytest.fixture
def tpath_preclear(request):
    """
    Fixture that indicates that the test path should be cleared automatically
    before running each test. This cleans up the test data.
    """
    tpath_raw = tpath_raw_make(request)
    no_preclear = request.config.getvalue('--no-preclear')
    if not no_preclear:
        rmtree(tpath_raw, ignore_errors = True)


@pytest.fixture
def tpath(request):
    """
    Fixture that takes the value of the special test-specific folder for test
    run data and plots. Usually the <folder of the test>/tresults/test_name/
    """
    tpath_raw = tpath_raw_make(request)

    os.makedirs(tpath_raw, exist_ok = True)
    os.utime(tpath_raw, None)
    return tpath_raw


@pytest.fixture
def tpath_join(request):
    """
    Fixture that joins subpaths to the value of the special test-specific folder for test
    run data and plots. Usually the <folder of the test>/tresults/test_name/.

    This function should be use like test_thing.save(tpath_join('output_file.png'))
    """
    tpath_raw = tpath_raw_make(request)
    first_call = True

    def tpath_joiner(*subpath):
        nonlocal first_call
        if first_call:
            os.makedirs(tpath_raw, exist_ok = True)
            os.utime(tpath_raw, None)
            first_call = False
        return path.join(tpath_raw, *subpath)

    return tpath_joiner


@pytest.fixture
def fpath(request):
    """
    py.test fixture that returns the folder path of the test being run. Useful
    for accessing data files.
    """
    return fpath_raw_make(request)


@pytest.fixture
def fpath_join(request):
    """
    py.test fixture that runs :code:`os.path.join(path, *arguments)` to merge subpaths
    with the folder path of the current test being run. Useful for referring to
    data files.
    """
    def join_func(*path):
        return os.path.join(fpath_raw_make(request), *path)
    return join_func

@pytest.fixture
def closefigs():
    import matplotlib.pyplot as plt
    yield
    plt.close('all')


@pytest.fixture
def test_trigger():
    """
    This fixture provides a contextmanager that causes a function to call
    if an AssertionError is raised. It will also call if any of its argument,
    or keyword arguments is true. This allows you to conveniently force
    calling using other flags or fixtures.

    The primary usage of this is to plot outputs only on test failures, while also
    allowing plotting to happen using the plot fixture and pytest cmdline argument
    """
    run_store = []

    @contextlib.contextmanager
    def fail(call, **kwargs):
        run_store.append(call)

        def call(did_fail):
            do_call = did_fail
            for k, v in kwargs.items():
                if v:
                    do_call = True
                    break

            if do_call:
                for call in run_store:
                    call(fail = did_fail, **kwargs)
                run_store.clear()
        try:
            yield
        except AssertionError:
            call(True)
            raise
        else:
            call(False)

        return
    return fail


@pytest.fixture()
def ic():
    """
    Fixture to provide icecream imports without requiring that the package exist
    """
    try:
        from icecream import ic
        return ic
    except ImportError:
        pass
    try:
        from IPython.lib.pretty import pprint
        return pprint
    except ImportError:
        from pprint import pprint
        return pprint


#these are used with the pprint fixture
try:
    import icecream
except ImportError:
    icecream = None
    pass
try:
    from IPython.lib.pretty import pprint, pretty
    pformat = pretty
except ImportError:
    from pprint import pprint, pformat


@pytest.fixture
def pprint(request, tpath_join):
    """
    This is a fixture providing a wrapper function for pretty printing. It uses
    the icecream module for pretty printing, falling back to ipythons pretty
    printer if needed, then to the python build in pretty printing module.

    Along with printing to stdout, this function prints into the tpath_folder to
    save all output into output.txt.
    """
    fname = tpath_join('output.txt')

    #pushes past the dot
    print('---------------:{}:--------------'.format(request.node.name))
    with open(fname, 'w') as F:
        def pprint(*args, F = F, pretty = True, **kwargs):
            outs = []
            if pretty:
                for arg in args:
                    outs.append(
                        pformat(arg)
                    )
            else:
                outs = args
            if F is not None:
                print(*outs, file = F)
            if icecream is not None:
                icecream.DEFAULT_OUTPUT_FUNCTION(' '.join(outs), **kwargs)
            else:
                print(*outs, **kwargs)

        yield pprint


def tpath_raw_make(request):
    if isinstance(request.node, pytest.Function):
        return relfile_test(request.node.function.__code__.co_filename, request, 'tresults')
    raise RuntimeError("TPath currently only works for functions")


def fpath_raw_make(request):
    if isinstance(request.node, pytest.Function):
        return os.path.split(request.node.function.__code__.co_filename)[0]
    raise RuntimeError("TPath currently only works for functions")


def relfile(_file_, *args, fname = None):
    fpath = path.split(_file_)[0]
    post = path.join(*args)
    fpath = path.join(fpath, post)
    #os.makedirs(fpath, exist_ok = True)
    #os.utime(fpath, None)

    if fname is None:
        return fpath
    else:
        return path.join(fpath, fname)


def relfile_test(_file_, request, pre = None, post = None, fname = None):
    """
    Generates a folder specific to py.test function
    (provided by using the "request" fixture in the test's arguments)
    """
    if isinstance(pre, (list, tuple)):
        pre = path.join(pre)

    testname = request.node.name
    if pre is not None:
        testname = path.join(pre, testname)

    if isinstance(post, (list, tuple)):
        post = path.join(post)
    if post is not None:
        return relfile(_file_, testname, post, fname = fname)
    else:
        return relfile(_file_, testname, fname = fname)




'''Functions to calculate Newtonian noise

'''
from __future__ import division
import numpy as np
from numpy import pi, sqrt, exp, log10
import scipy.integrate as scint
import scipy.special as sp

from .seismic import seismic_ground_NLNM
from .. import const


def gravg(f, seismic):
    """Gravity gradient noise for single test mass

    :f: frequency array in Hz
    :seismic: gwinc Seismic struct

    :returns: displacement noise power spectrum at :f:, in meters

    References:

     Saulson 1984,           http://dx.doi.org/10.1103/PhysRevD.30.732
     Hughes and Thorne 1998, http://dx.doi.org/10.1103/PhysRevD.58.122002

     Driggers and Harms 2011, ``Results of Phase 1 Newtonian Noise
     Measurements at the LIGO Sites,'' February-March 2011.  T1100237.
     https://dcc.ligo.org/LIGO-T1100237

    Written by Enrico Camagna (?)

    added to Bench by Gregg Harry 8/27/03
    seismic spectrum modified by Jan Harms 05/11/2010
    Calculates gravity gradient noise for four mirrors

    """

    fk = seismic.KneeFrequency
    a = seismic.LowFrequencyLevel
    gamma = seismic.Gamma
    rho = seismic.Rho
    # factor to account for correlation between masses
    # and the height of the mirror above the ground
    beta = seismic.Beta
    h = seismic.TestMassHeight
    c_rayleigh = seismic.RayleighWaveSpeed

    if 'Omicron' in seismic:
        omicron = seismic.Omicron
    else:
        omicron = 1

    # a sort of theta function (Fermi distr.)
    coeff = 3**(-gamma*f)/(3**(-gamma*f) + 3**(-gamma*fk))

    # modelization of seismic noise (vertical)
    ground = a*coeff + a*(1-coeff)*(fk/f)**2
    if 'Site' in seismic and seismic.Site == 'LLO':
        ground = a*coeff*(fk/f) + a*(1-coeff)*(fk/f)**2

    # effective GG spring frequency, with G gravitational
    fgg = sqrt(const.G * rho) / (2*pi)

    # fixed numerical factors, 5/9/06, PF
    n = (beta*2*pi*(fgg**2/f**2)*ground)**2

    # The following two terms are corrections due to Jan Harms
    # https://git.ligo.org/rana-adhikari/CryogenicLIGO/issues/45
    # (1) projection of NN force onto the direction of the arm
    n = n * 1/2
    # (2) exponential cutoff at frequency (seismic speed)/(test mass height)
    n = n * exp(-4*pi*f*h/c_rayleigh)

    # Feedforward cancellation
    n /= (omicron**2)

    return n


def gravg_rayleigh(f, seismic):
    """Gravity gradient noise for single arm cavity from seismic Rayleigh waves

    :f: frequency array in Hz
    :seismic: gwinc Seismic structure

    :returns: displacement noise power spectrum at :f:, in meters

    Following Harms LRR: https://doi.org/10.1007/lrr-2015-3
    and Amann et al.: https://doi.org/10.1063/5.0018414

    """
    fk = seismic.KneeFrequency
    a = seismic.LowFrequencyLevel
    gamma = seismic.Gamma
    rho = seismic.Rho
    h = seismic.TestMassHeight
    c_rayleigh = seismic.RayleighWaveSpeed

    if 'Omicron' in seismic:
        omicron = seismic.Omicron
    else:
        omicron = 1

    # a sort of theta function (Fermi distr.)
    coeff = 3**(-gamma*f)/(3**(-gamma*f) + 3**(-gamma*fk))

    # modelization of seismic noise (vertical)
    ground = a*coeff + a*(1-coeff)*(fk/f)**2
    if 'Site' in seismic and seismic.Site == 'LLO':
        ground = a*coeff*(fk/f) + a*(1-coeff)*(fk/f)**2

    # Harms LRR eqs. 35, 96, and 98
    w = 2 * pi * f
    k = w / c_rayleigh
    kP = w / seismic.pWaveSpeed
    kS = w / seismic.sWaveSpeed
    qzP = sqrt(k**2 - kP**2)
    qzS = sqrt(k**2 - kS**2)
    zeta = sqrt(qzP / qzS)

    gnu = k * (1 - zeta) / (qzP - k * zeta)
    
    if h >= 0:
        # Harms LRR
        n = (2 * pi * const.G * rho * exp(-h * k) * gnu)**2 * ground**2 / w**4

    else:
        # Amann et al., eqs. 2-6. Note h is there defined as depth;
        # We define it as height.
        r0 = k * (1 - zeta)
        sh = -k * (1 + zeta) * exp(h * k)
        bh = (2 / 3) * (2 * k * exp(h * qzP) + zeta * qzS * exp(h * qzS))
        Rcal = np.abs((sh + bh) / r0)**2
        n = 4 * (sqrt(2) * pi * const.G * rho * gnu)**2 * Rcal * ground**2 / w**4

    n /= omicron**2

    return n


def gravg_pwave(f, seismic, exact=False):
    """Gravity gradient noise for single test mass from seismic p-waves

    :f: frequency array in Hz
    :seismic: gwinc Seismic structure
    :exact: whether to use a slower numerical integral good to higher frequencies
            or faster special functions. If exact=False, nan is returned at
            high frequencies where the special functions have numerical errors.

    :returns: displacement noise power spectrum at :f:, in meters

    Following Harms LRR: https://doi.org/10.1007/lrr-2015-3

    """
    cP = seismic.pWaveSpeed
    levelP = seismic.pWaveLevel
    tmheight = seismic.TestMassHeight
    rho_ground = seismic.Rho

    kP = (2 * pi * f) / cP

    psd_ground_pwave = (levelP * seismic_ground_NLNM(f))**2

    xP = np.abs(kP * tmheight)

    if tmheight >= 0:
        # Surface facility
        # The P-S conversion at the surface is not implemented
        if exact:
            height_supp_power = (3 / 2) * np.array(
                [scint.quad(lambda th, x: np.sin(th)**3
                            * np.exp(-2 * x * np.sin(th)), 0, pi / 2, args=(x,))[0]
                 for x in xP])

        else:
            xP[xP > 10] = np.nan
            height_supp_power = 3 / (24*xP) * (
                8*xP - 9*pi*sp.iv(2, 2*xP) - 6*pi*xP*sp.iv(3, 2*xP)
                + 6*pi*xP*sp.modstruve(1, 2*xP) - 3*pi*sp.modstruve(2, 2*xP))

    else:
        # Underground facility
        # The cavity effect is not included
        if exact:
            height_supp_power = (3 / 4) * np.array(
                [scint.quad(lambda th, x: np.sin(th)**3
                            * (2 - np.exp(-x * np.sin(th)))**2, 0, pi, args=(x,))[0]
                 for x in xP])

        else:
            xP[xP > 2] = np.nan
            height_supp_power = 1 + 3*pi/(8*xP) * (
                24*sp.iv(2, xP) - 3*sp.iv(2, 2*xP) + 8*xP*sp.iv(3, xP)
                - 2*xP*sp.iv(3, 2*xP) - 8*xP*sp.modstruve(1, xP)
                + 2*xP*sp.modstruve(1, 2*xP) - 8*sp.modstruve(2, xP)
                + sp.modstruve(2, 2*xP))

    psd_gravg_pwave = ((2 * pi * const.G * rho_ground)**2
            * psd_ground_pwave * height_supp_power)
    psd_gravg_pwave /= (2 * pi * f)**4
    return psd_gravg_pwave


def gravg_swave(f, seismic, exact=False):
    """Gravity gradient noise for single test mass from seismic s-waves

    :f: frequency array in Hz
    :seismic: gwinc Seismic structure
    :exact: whether to use a slower numerical integral good to higher frequencies
            or faster special functions. If exact=False, nan is returned at
            high frequencies where the special functions have numerical errors.

    :returns: displacement noise power spectrum at :f:, in meters

    Following Harms LRR: https://doi.org/10.1007/lrr-2015-3

    """
    cS = seismic.sWaveSpeed
    levelS = seismic.sWaveLevel
    tmheight = seismic.TestMassHeight
    rho_ground = seismic.Rho

    kS = (2 * pi * f) / cS

    psd_ground_swave = (levelS * seismic_ground_NLNM(f))**2

    xS = np.abs(kS * tmheight)

    # For both surface and underground facilities
    if exact:
        height_supp_power = (3 / 2) * np.array(
            [scint.quad(lambda th, x: np.sin(th)**3
                        * np.exp(-2 * x * np.sin(th)), 0, pi / 2, args=(x,))[0]
             for x in xS])

    else:
        xS[xS > 10] = np.nan
        height_supp_power = 3 / (24*xS) * (
            8*xS - 9*pi*sp.iv(2, 2*xS) - 6*pi*xS*sp.iv(3, 2*xS)
            + 6*pi*xS*sp.modstruve(1, 2*xS) - 3*pi*sp.modstruve(2, 2*xS))

    psd_gravg_swave = ((2 * pi * const.G * rho_ground)**2
            * psd_ground_swave * height_supp_power)
    psd_gravg_swave /= (2 * pi * f)**4

    return psd_gravg_swave


def atmois(f, atmos, seismic):
    """Atmospheric infrasound newtonian noise for single arm cavity

    :f: frequency array in Hz
    :atmos: gwinc Atmospheric structure
    :seismic: gwinc Seismic structure

    :returns: displacement noise power spectrum at :f:, in meters

    """
    p_air = atmos.AirPressure
    rho_air = atmos.AirDensity
    ai_air = atmos.AdiabaticIndex
    c_sound = atmos.SoundSpeed
    h = seismic.TestMassHeight

    w = 2 * pi * f
    k = w / c_sound

    # Pressure spectrum
    try:
        a_if = atmos.InfrasoundLevel1Hz
        e_if = atmos.InfrasoundExponent
        psd_if = (a_if * f**e_if)**2
    except AttributeError:
        psd_if = atmoBowman(f)**2

    # Harms LRR (2015), eq. 172
    # https://doi.org/10.1007/lrr-2015-3
    # And with the Bessel terms ignored... for 4 km this amounts to a 10%
    # correction at 10 Hz and a 30% correction at 1 Hz
    coupling_if = 2/3 * (4 * pi / (k * w**2) * const.G * rho_air / (ai_air * p_air))**2

    n_if = coupling_if * psd_if

    return n_if


def atmoBowman(f):
    """The Bowman infrasound model
   
    """ 
    freq = np.array([
        0.01, 0.0155, 0.0239, 0.0367, 0.0567,
        0.0874, 0.1345, 0.2075, 0.32, 0.5,
        0.76, 1.17, 1.8, 2.79, 4.3,
        6.64, 10, 100,
    ])
    pressure_asd = sqrt([
        22.8, 4, 0.7, 0.14, 0.027, 0.004,
        0.0029, 0.0039, 7e-4, 1.44e-4, 0.37e-4,
        0.12e-4, 0.56e-5, 0.35e-5, 0.26e-5, 0.24e-5,
        2e-6, 2e-6,
    ])
    return 10**(np.interp(log10(f), log10(freq), log10(pressure_asd)))



import numpy as np

from .. import freq_from_spec, load_budget
from .. import nb


def test_nb():
    freq = freq_from_spec()
    budget = load_budget('Aplus', freq)
    assert np.all(budget.freq == freq)



from __future__ import print_function
import os
import signal
import logging
import argparse

from . import (
    __version__,
    IFOS,
    DEFAULT_FREQ,
    InvalidFrequencySpec,
    load_budget,
    logger,
)
from . import io

logger.setLevel(os.getenv('LOG_LEVEL', 'WARNING').upper())
formatter = logging.Formatter('%(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)

##################################################

description = """GWINC noise budget tool

IFOs can be specified by name of included canonical budget (see
below), or by path to a budget module (.py), description file
(.yaml/.mat/.m), or HDF5 data file (.hdf5/.h5).  Available included
IFOs are:

  {}

""".format(', '.join(["{}".format(ifo) for ifo in IFOS]))
# for ifo in available_ifos():
#     description += "  '{}'\n".format(ifo)
description += """

By default the noise budget of the specified IFO will be loaded and
plotted with an interactive plotter.  Individual IFO parameters can be
overriden with the --ifo option:

  gwinc --ifo Optics.SRM.Tunephase=3.14 ...

If the --save option is specified the plot will be saved directly to a
file (without display) (various file formats are supported, indicated
by file extension).  If the requested extension is 'hdf5' or 'h5' then
the noise traces and IFO parameters will be saved to an HDF5 file.

If the --range option is specified and the inspiral_range package is
available, various BNS (m1=m2=1.4 M_solar) range figures of merit will
be calculated for the resultant spectrum.  The default waveform
parameters can be overriden with the --waveform-parameter/-wp option:

  gwinc -r -wp m1=20 -wp m2=20 ...

See the inspiral_range package documentation for details.
"""

IFO = 'aLIGO'
RANGE_PARAMS = dict(m1=1.4, m2=1.4)
DATA_SAVE_FORMATS = ['.hdf5', '.h5']

parser = argparse.ArgumentParser(
    prog='gwinc',
    description=description,
    formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument(
    '--version', '-v', action='version', version=__version__)
parser.add_argument(
    '--freq', '-f', metavar='FLO:[NPOINTS:]FHI',
    help="logarithmic frequency array specification in Hz [{}]".format(DEFAULT_FREQ))
parser.add_argument(
    '--ifo', '-o', metavar='PARAM=VAL', default=[],
    #nargs='+', action='extend',
    action='append',
    help="override budget IFO parameter (may be specified multiple times)")
parser.add_argument(
    '--title', '-t',
    help="plot title")
parser.add_argument(
    '--range', '-r', action='store_true',
    help="calculate inspiral ranges [m1=m2=1.4]")
parser.add_argument(
    '--waveform-parameter', '-wp', metavar='PARAM=VAL', default=[],
    action='append',
    help="specify inspiral range parameters (may be specified multiple times)")
group = parser.add_mutually_exclusive_group()
group.add_argument(
    '--interactive', '-i', action='store_true',
    help="launch interactive shell after budget processing")
group.add_argument(
    '--save', '-s', metavar='PATH', action='append',
    help="save plot (.png/.pdf/.svg) or budget traces (.hdf5/.h5) to file (may be specified multiple times)")
group.add_argument(
    '--yaml', '-y', action='store_true',
    help="print IFO as yaml to stdout and exit (budget not calculated)")
group.add_argument(
    '--text', '-x', action='store_true',
    help="print IFO as text table to stdout and exit (budget not calculated)")
group.add_argument(
    '--diff', '-d', metavar='IFO',
    help="show difference table between IFO and another IFO description (name or path) and exit (budget not calculated)")
group.add_argument(
    '--list', '-l', action='store_true',
    help="list all elements of Budget (budget not calculated)")
parser.add_argument(
    '--no-plot', '-np', action='store_false', dest='plot',
    help="suppress plotting")
parser.add_argument(
    '--bname', '-b',
    help="name of top-level Budget class to load (defaults to IFO name)")
parser.add_argument(
    'IFO',
    help="IFO name or path")
parser.add_argument(
    'subbudget', metavar='SUBBUDGET', nargs='?',
    help="subbudget to plot; can be nested (e.g. 'Thermal.Substrate')")


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    args = parser.parse_args()

    ##########
    # initial arg processing

    if os.path.splitext(os.path.basename(args.IFO))[1] in io.DATA_SAVE_FORMATS:
        if args.freq:
            parser.exit(2, "Error: Frequency specification not allowed when loading traces from file.\n")
        if args.ifo:
            parser.exit(2, "Error: IFO parameter specification not allowed when loading traces from file.\n")
        from .io import load_hdf5
        budget = None
        name = args.IFO
        trace = load_hdf5(args.IFO)
        freq = trace.freq
        ifo = trace.ifo
        plot_style = trace.plot_style

    else:
        try:
            budget = load_budget(args.IFO, freq=args.freq, bname=args.bname)
        except InvalidFrequencySpec as e:
            parser.error(e)
        except RuntimeError as e:
            parser.exit(2, f"Error: {e}\n")
        name = budget.name
        ifo = budget.ifo
        freq = budget.freq
        plot_style = getattr(budget, 'plot_style', {})
        trace = None

    for paramval in args.ifo:
        try:
            param, val = paramval.split('=', 1)
            ifo[param] = float(val)
        except ValueError:
            parser.error(f"Improper IFO parameter specification: {paramval}")

    if args.yaml:
        if not ifo:
            parser.exit(2, "Error: IFO structure not provided.\n")
        print(ifo.to_yaml(), end='')
        return
    if args.text:
        if not ifo:
            parser.exit(2, "Error: IFO structure not provided.\n")
        print(ifo.to_txt(), end='')
        return
    if args.diff:
        if not ifo:
            parser.exit(2, "Error: IFO structure not provided.\n")
        dbudget = load_budget(args.diff)
        diffs = ifo.diff(dbudget.ifo)
        if diffs:
            w = max([len(d[0]) for d in diffs])
            fmt = '{{:{}}} {{:>20}} {{:>20}}'.format(w)
            print(fmt.format('', args.IFO, args.diff))
            print(fmt.format('', '-----', '-----'))
            for p in diffs:
                k = str(p[0])
                v = repr(p[1])
                ov = repr(p[2])
                print(fmt.format(k, v, ov))
        return
    if args.list:
        for i in budget.walk():
            name = '.'.join([n.__class__.__name__ for n in i])
            type = i[-1].__class__.__bases__[0].__name__
            print(f'{name} ({type})')
        return

    if args.subbudget:
        try:
            budget[args.subbudget]
        except KeyError:
            parser.exit(3, f"Error: Unknown budget item '{args.subbudget}'.\n")

    out_data_files = set()
    out_plot_files = set()
    if args.save:
        args.plot = False
        out_files = set(args.save)
        for path in out_files:
            if os.path.splitext(path)[1] in io.DATA_SAVE_FORMATS:
                out_data_files.add(path)
        out_plot_files = out_files - out_data_files

    if args.plot or out_plot_files:
        if out_plot_files:
            # FIXME: this silliness seems to be the only way to have
            # matplotlib usable on systems without a display.  There must
            # be a better way.  'AGG' is a backend that works without
            # displays.  but it has to be set before any other matplotlib
            # stuff is imported.  and we *don't* want it set if we do want
            # to show an interactive plot.  there doesn't seem a way to
            # set this opportunistically.
            import matplotlib
            matplotlib.use('AGG')
        try:
            from matplotlib import pyplot as plt
        except ImportError as e:
            parser.exit(5, f"ImportError: {e}\n")
        except RuntimeError:
            parser.exit(10, "Error: Could not open display for plotting.\n")

    if args.range:
        try:
            import inspiral_range
        except ImportError as e:
            parser.exit(5, f"ImportError: {e}\n")

        logger_ir = logging.getLogger('inspiral_range')
        logger_ir.setLevel(logger.getEffectiveLevel())
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(name)s: %(message)s'))
        logger_ir.addHandler(handler)

        for paramval in args.waveform_parameter:
            try:
                param, val = paramval.split('=')
                if not val:
                    raise ValueError
            except ValueError:
                parser.error(f"Improper range parameter specification: {paramval}")
            try:
                val = float(val)
            except ValueError:
                pass
            RANGE_PARAMS[param] = val

    ##########
    # main calculations

    if not trace:
        logger.info("calculating budget...")
        trace = budget.run()

    if args.range:
        logger.info("calculating inspiral ranges...")
        metrics, H = inspiral_range.all_ranges(freq, trace.psd, **RANGE_PARAMS)
        print(f"{H.params['approximant']} {H.params['m1']}/{H.params['m2']} M_solar:")
        for metric, (value, unit) in metrics.items():
            if unit is None:
                unit = ''
            print(f" {metric}: {value:0.1f} {unit}")
        range_func = 'range'
        subtitle = 'inspiral {func} {m1}/{m2} $\mathrm{{M}}_\odot$: {fom:.0f} {unit}'.format(
            func=range_func,
            m1=H.params['m1'],
            m2=H.params['m2'],
            fom=metrics[range_func][0],
            unit=metrics[range_func][1] or '',
        )
    else:
        subtitle = None

    if args.subbudget:
        trace = trace[args.subbudget]
        name += f': {args.subbudget}'

    if args.title:
        plot_style['title'] = args.title
    else:
        plot_style['title'] = "GWINC Noise Budget: {}".format(name)

    ##########
    # interactive

    # interactive shell plotting
    if args.interactive:
        banner = """GWINC interactive shell

The 'ifo' Struct, 'budget', and 'trace' objects are available for
inspection.  Use the 'whos' command to view the workspace.
"""
        if not args.plot:
            banner += """
You may plot the budget using the 'trace.plot()' method:

In [.]: trace.plot(**plot_style)
"""
        banner += """
You may interact with the plot using the 'plt' functions, e.g.:

In [.]: plt.title("foo")
In [.]: plt.savefig("foo.pdf")
"""
        from IPython.terminal.embed import InteractiveShellEmbed
        if subtitle:
            plot_style['title'] += '\n' + subtitle
        ipshell = InteractiveShellEmbed(
            banner1=banner,
            user_ns={
                'ifo': ifo,
                'budget': budget,
                'trace': trace,
                'plot_style': plot_style,
            },
        )
        ipshell.enable_pylab(import_all=False)
        if args.plot:
            ipshell.ex("fig = trace.plot(**plot_style)")
        ipshell()

    ##########
    # output

    # save noise trace to HDF5 file
    if out_data_files:
        for path in out_data_files:
            logger.info("saving budget trace: {}".format(path))
            io.save_hdf5(
                trace=trace,
                path=path,
                ifo=ifo,
                plot_style=plot_style,
            )

    # standard plotting
    if args.plot or out_plot_files:
        logger.debug("plotting noises...")
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        if subtitle:
            plot_style['title'] += '\n' + subtitle
        trace.plot(
            ax=ax,
            **plot_style
        )
        fig.tight_layout()
        if out_plot_files:
            for path in out_plot_files:
                logger.info("saving budget plot: {}".format(path))
                try:
                    fig.savefig(path)
                except Exception as e:
                    parser.exit(2, f"Error saving plot: {e}.\n")
        else:
            plt.show()


if __name__ == '__main__':
    main()



'''Functions to calculate seismic noise in suspended optics.

'''
from __future__ import division
import numpy as np
from scipy.interpolate import PchipInterpolator as interp1d


def seismic_suspension_filtered(sustf, in_trans, direction):
    """Seismic displacement noise for single suspended test mass.

    :sustf: sus transfer function Struct
    :in_trans: input translational displacement spectrum
    :direction: 'horiz' for horizontal or 'vert' for vertical

    :returns: tuple of displacement noise power spectrum at :f:

    """
    if direction == 'horiz':
        # horizontal noise total
        n = (abs(sustf.hTable)**2) * in_trans**2

    elif direction == 'vert':
        # vertical to horizontal coupling
        theta = sustf.VHCoupling.theta

        # vertical noise total
        n = (abs(theta * sustf.vTable)**2) * in_trans**2

    return n


def platform_motion(f, ifo):
    """Compute the platform motion

    :f: frequency array in Hz
    :ifo: the IFO struct

    :returns: tuple of displacement noise power spectrum at :f: for
    translational and rotational DOFs.
    """
    if 'PlatformMotion' in ifo.Seismic:
        if ifo.Seismic.PlatformMotion == 'BSC':
            nt, nr = seismic_BSC_ISI(f)
        elif ifo.Seismic.PlatformMotion == '6D':
            nt, nr = seismic_BSC_ISI_6D(f)
        elif ifo.Seismic.PlatformMotion == 'intermediate':
            nt_isi, nr_isi = seismic_BSC_ISI(f)
            nt_6d, nr_6d = seismic_BSC_ISI_6D(f)
            nt = np.sqrt(nt_isi * nt_6d)
            nr = np.sqrt(nr_isi * nr_6d)
        else:
            raise ValueError(
                'Unrecognized platform motion ' + ifo.Seismic.PlatformMotion)
    else:
        nt, nr = seismic_BSC_ISI(f)

    return nt, nr


def seismic_BSC_ISI(f):
    """Rough seismic noise spectra on aLIGO BSC ISI table.

    :f: frequency array in Hz

    :returns: tuple of displacement noise power spectrum at :f: for
    translational and rotational DOFs.

    """
    SEI_F = np.array([0.01, 0.03, 0.1, 0.2, 0.5, 1, 10, 30, 300])

    # translational DOFs
    SEI_T = np.array([3e-6, 1e-6, 2e-7, 2e-7, 8e-10, 1e-11, 3e-13, 3e-14, 3e-14])
    nt = 10**(interp1d(SEI_F, np.log10(SEI_T))(f))

    # rotational DOFs
    SEI_R = np.array([1e-8, 3e-8, 2e-8, 1e-8, 4e-10, 1e-11, 3e-13, 3e-14, 3e-14])
    nr = 10**(interp1d(SEI_F, np.log10(SEI_R))(f))

    return nt, nr


def seismic_BSC_ISI_6D(f):
    """Rough seismic noise spectra on aLIGO BSC ISI table with a 6D seismometer.

    This largely follows Mow-Lowry and Martynov, arXiv:1801.01468.

    :f: frequency array in Hz

    :returns: tuple of displacement noise power spectrum at :f: for
    translational and rotational DOFs.

    """
    # FIXME: merge this with above, using flag

    SEI_F = np.array([0.01, 0.03, 0.1, 0.2, 0.5, 1, 10, 100, 300])

    SEI_T_self = np.array([1e-7, 1e-9, 3e-11, 6e-12, 3e-13, 1e-13, 3e-14, 1e-14, 1e-14])
    nt_self = 10**(interp1d(SEI_F, np.log10(SEI_T_self))(f))
    nt_gnd = 10*seismic_ground_NLNM(f)
    blend_t = np.abs(100/(1+1j*f/0.01)**4)
    nt = np.sqrt(nt_self**2 + (blend_t * nt_gnd)**2)

    SEI_R_self = np.array([2e-11, 5e-12, 1e-12, 6e-13, 3e-13, 2e-13, 6e-14, 2e-14, 2e-14])
    nr_self = 10**(interp1d(SEI_F, np.log10(SEI_R_self))(f))
    nr_gnd = np.abs(1e-7/(1+1j*f/0.001))
    blend_r = np.abs(100/(1+1j*f/0.01)**4)
    nr = np.sqrt(nr_self**2 + (blend_r * nr_gnd)**2)

    return nt, nr


def seismic_ground_NLNM(f):
    """The Peterson new generic ground motion low noise model.

    :f: frequency array in Hz

    :returns: displacement noise amplitude spectrum at :f:

    """
    Pl = np.array([
        1.00e-02, 1.00e-01, 1.70e-01, 4.00e-01, 8.00e-01, 1.24e+00,
        2.40e+00, 4.30e+00, 5.00e+00, 6.00e+00, 1.00e+01, 1.20e+01,
        1.56e+01, 2.19e+01, 3.16e+01, 4.50e+01, 7.00e+01, 1.01e+02,
        1.54e+02, 3.28e+02, 6.00e+02, 1.00e+04])
    Al = np.array([
        -156.72, -162.36, -166.7, -170.0, -166.4, -168.6, -159.98,
        -141.1, -71.36, -97.26, -132.18, -205.27, -37.65, -114.37,
        -160.58, -187.5, -216.47, -185.0, -168.34, -217.43, -258.28,
        -346.88])
    Bl = np.array([
        5.64, 5.64, 0.0, -8.3, 28.9, 52.48, 29.81,
        0.0, -99.77, -66.49, -31.57, 36.16, -104.33, -47.1,
        -16.28, 0.0, 15.7, 0.0, -7.61, 11.9, 26.6,
        48.75])
    nlnm = 10**(np.interp(1/f, Pl, Al+Bl*np.log10(Pl))/20) / (2 * np.pi * f)**2
    return nlnm


def seismic_ground_NHNM(f):
    """The Peterson new generic ground motion high noise model.

    :f: frequency array in Hz

    :returns: displacement noise amplitude spectrum at :f:

    """

    Pl = np.array([
        1.00e-01, 2.20e-01, 3.20e-01, 8.00e-01, 3.80e+00,
        4.60e+00, 6.30e+00, 7.90e+00, 1.54e+01, 2.00e+01,
        3.54e+02,
        ])
    Al = np.array([
        -108.73, -150.34, -122.31, -116.85, -108.48,
        -74.66, 0.66, -93.37, 73.54, -151.52,
        -206.66,
        ])
    Bl = np.array([
        -17.23, -80.50, -23.87, 32.51, 18.08,
        -32.95, -127.18, -22.42, -162.98, 10.01,
        31.63,
        ])
    nhnm = 10**(np.interp(1/f, Pl, Al+Bl*np.log10(Pl))/20) / (2 * np.pi * f)**2
    return nhnm



'''Functions to calculate substrate thermal noise

'''
from __future__ import division, print_function
import numpy as np
from numpy import exp, inf, pi, sqrt
import scipy.special
import scipy.integrate

from .. import const
from ..const import BESSEL_ZEROS as zeta
from ..const import J0M as j0m


def substrate_thermorefractive(f, materials, wBeam, exact=False):
    """Substrate thermal displacement noise spectrum from thermorefractive fluctuations

    :f: frequency array in Hz
    :materials: gwinc optic materials structure
    :wBeam: beam radius (at 1 / e^2 power)
    :exact: whether to use adiabatic approximation or exact calculation (False)

    :returns: displacement noise power spectrum at :f:, in meters

    """
    H = materials.MassThickness
    kBT = const.kB * materials.Substrate.Temp
    Temp = materials.Substrate.Temp
    rho = materials.Substrate.MassDensity
    beta = materials.Substrate.dndT
    C = materials.Substrate.MassCM
    kappa = materials.Substrate.MassKappa
    r0 = wBeam/np.sqrt(2)
    omega = 2*pi*f

    if exact:
        # arXiv:cond-mat/0402650, Eq. E7
        w = omega * r0**2 * rho * C / (2 * kappa)
        psd = np.abs(H * beta**2 * kBT * Temp / (2 * pi * kappa) * (exp(1j*w) * scipy.special.exp1(1j*w)
            + exp(-1j*w) * scipy.special.exp1(-1j*w)))

    else:
        # arXiv:cond-mat/0402650, Eq. 5.3; P1400084, Eq. 18
        psd = 4*H*beta**2*kappa*kBT*Temp/(pi*r0**4*omega**2*(rho*C)**2)

    return psd


def substrate_brownian(f, materials, wBeam):
    """Substrate thermal displacement noise spectrum due to substrate mechanical loss

    :f: frequency array in Hz
    :materials: gwinc optic materials structure
    :wBeam: beam radius (at 1 / e^2 power)

    :returns: displacement noise power spectrum at :f:, in meters

    """
    Y = materials.Substrate.MirrorY
    sigma = materials.Substrate.MirrorSigma
    c2 = materials.Substrate.c2
    n = materials.Substrate.MechanicalLossExponent
    alphas = materials.Substrate.Alphas
    kBT = const.kB * materials.Substrate.Temp

    cftm, aftm = substrate_brownian_FiniteCorr(materials, wBeam)

    # Bulk substrate contribution
    phibulk = c2 * f**n
    cbulk = 8 * kBT * aftm * phibulk / (2 * pi * f)

    # Surface loss contribution
    # csurf = alphas/(Y*pi*wBeam^2)
    csurf = alphas*(1-2*sigma)/((1-sigma)*Y*pi*wBeam**2)
    csurf *= 8 * kBT / (2 * pi * f)

    return csurf + cbulk


def substrate_brownian_FiniteCorr(materials, wBeam):
    """Substrate brownian noise finite-size test mass correction

    :materials: gwinc optic materials structure
    :wBeam: beam radius (at 1 / e^2 power)

    :returns: correction factors tuple:
    cftm = finite mirror correction factor
    aftm = amplitude coefficient for thermal noise:
           thermal noise contribution to displacement noise is
           S_x(f) = (8 * kB * T / (2*pi*f)) * Phi(f) * aftm

    Equation references to Bondu, et al. Physics Letters A 246 (1998)
    227-236 (hereafter BHV) and Liu and Thorne gr-qc/0002055 (hereafter LT)

    """
    a = materials.MassRadius
    h = materials.MassThickness
    Y = materials.Substrate.MirrorY
    sigma = materials.Substrate.MirrorSigma

    # LT uses e-folding of power
    r0 = wBeam / sqrt(2)
    km = zeta/a

    Qm = exp(-2*km*h) # LT eq. 35a

    Um = (1-Qm)*(1+Qm)+4*h*km*Qm
    Um = Um/((1-Qm)**2-4*(km*h)**2*Qm) # LT 53 (BHV eq. btwn 29 & 30)

    x = exp(-(zeta*r0/a)**2/4)
    s = sum(x/(zeta**2*j0m)) # LT 57

    x2 = x*x
    U0 = sum(Um*x2/(zeta*j0m**2))
    U0 = U0*(1-sigma)*(1+sigma)/(pi*a*Y) # LT 56 (BHV eq. 3)

    p0 = 1/(pi*a**2) # LT 28
    DeltaU = (pi*h**2*p0)**2
    DeltaU = DeltaU + 12*pi*h**2*p0*sigma*s
    DeltaU = DeltaU + 72*(1-sigma)*s**2
    DeltaU = DeltaU*a**2/(6*pi*h**3*Y) # LT 54

    # LT 58 (eq. following BHV 31)
    aftm = DeltaU + U0

    # amplitude coef for infinite TM, LT 59
    # factored out: (8 * kB * T * Phi) / (2 * pi * f)
    aitm = (1 - sigma**2) / (2 * sqrt(2 * pi) * Y * r0)

    # finite mirror correction
    cftm = aftm / aitm

    return cftm, aftm


def substrate_thermoelastic(f, materials, wBeam):
    """Substrate thermal displacement noise spectrum from thermoelastic fluctuations

    :f: frequency array in Hz
    :materials: gwinc optic materials structure
    :wBeam: beam radius (at 1 / e^2 power)

    :returns: displacement noise power spectrum at :f:, in meters

    """
    sigma = materials.Substrate.MirrorSigma
    rho = materials.Substrate.MassDensity
    kappa = materials.Substrate.MassKappa # thermal conductivity
    alpha = materials.Substrate.MassAlpha # thermal expansion
    CM = materials.Substrate.MassCM # heat capacity @ constant mass
    Temp = materials.Substrate.Temp # temperature
    kBT = const.kB * materials.Substrate.Temp

    S = 8*(1+sigma)**2*kappa*alpha**2*Temp*kBT # note kBT has factor Temp
    S /= (sqrt(2*pi)*(CM*rho)**2)
    S /= (wBeam/sqrt(2))**3 # LT 18 less factor 1/omega^2

    # Corrections for finite test masses:
    S *= substrate_thermoelastic_FiniteCorr(materials, wBeam)

    return S/(2*pi*f)**2


def substrate_thermoelastic_FiniteCorr(materials, wBeam):
    """Substrate thermoelastic noise finite-size test mass correction

    :materials: gwinc optic materials structure
    :wBeam: beam radius (at 1 / e^2 power)

    :returns: correction factor

    (Liu & Thorne gr-qc/0002055 equation 46)

    Equation references to Bondu, et al. Physics Letters A 246 (1998)
    227-236 (hereafter BHV) or Liu and Thorne gr-qc/0002055 (hereafter LT)

    """
    a = materials.MassRadius
    h = materials.MassThickness
    sigma = materials.Substrate.MirrorSigma

    # LT uses power e-folding
    r0 = wBeam/sqrt(2)
    km = zeta/a

    Qm = exp(-2*km*h) # LT 35a

    pm = exp(-(km*r0)**2/4)/(pi*(a*j0m)**2) # LT 37

    c0 = 6*(a/h)**2*sum(j0m*pm/zeta**2) # LT 32
    c1 = -2*c0/h # LT 32
    p0 = 1/(pi*a**2) # LT 28
    c1 += p0/(2*h) # LT 40

    coeff = (1-Qm)*((1-Qm)*(1+Qm)+8*h*km*Qm)
    coeff += 4*(h*km)**2*Qm*(1+Qm)
    coeff *= km*(pm*j0m)**2*(1-Qm)
    coeff /= ((1-Qm)**2-4*(h*km)**2*Qm)**2
    coeff = sum(coeff) + h*c1**2/(1+sigma)**2
    coeff *= (sqrt(2*pi)*r0)**3*a**2 # LT 46

    return coeff



'''Functions to calculate suspension thermal noise

'''
from __future__ import division
import numpy as np
from numpy import pi, imag

from ..const import kB
from ..suspension import getJointParams


def getJointTempSusceptibility(sus, sustf, stage_num, isUpper, direction):
    """Get the product of the temperature and the susceptibility of a joint

    :sus: the suspension struct
    :sustf: suspension transfer function struct
    :stage_num: stage number with 0 at the top
    :isUpper: whether to get the upper or lower joint
    :direction: either 'horiz' or 'vert'
    """
    jointParams = getJointParams(sus, stage_num)
    for (isLower, Temp, wireMat, bladeMat) in jointParams:
        if isUpper == isLower:
            continue

        ind = 2*stage_num + isLower
        # Compute the force on the TM along the beamline
        if direction == 'horiz':
            dxdF = sustf.hForce[ind]

        elif direction == 'vert':
            # vertical to beamline coupling angle
            theta = sustf.VHCoupling.theta

            # convert to beam line motion. theta is squared because
            # we rotate by theta into the suspension basis, and by
            # theta to rotate back to the beam line basis
            dxdF = theta**2 * sustf.vForce[ind]

        return Temp * abs(imag(dxdF))


def susptherm_stage(f, sus, sustf, stage_num, direction):
    """Suspension thermal noise of a single stage

    :f: frequency array in Hz
    :sus: gwinc suspension struct
    :sustf: suspension transfer function struct
    :stage_num: stage number with 0 at the top
    :direction: either 'horiz' for horizontal or 'vert' for vertical
    :comp: component of the loss 0) bulk, 1) surface, 2) TE
    :returns: displacement noise power spectrum at :f:

    :Temp: must either be set for each stage individually, or globally
    in :sus.Temp:.  If both are specified, :sus.Temp: is interpreted as
    the temperature of the upper joint of the top stage.

    Assumes suspension transfer functions and V-H coupling have been
    pre-calculated and populated into the relevant `sus` fields.

    """

    w = 2*pi*f
    last_stage = len(sus.Stage) - 1

    # get the noise from the lower joint
    noise = getJointTempSusceptibility(sus, sustf, stage_num, False, direction)

    # if this is the top mass, also get the noise from the upper joint
    if stage_num == 0:
        noise += getJointTempSusceptibility(sus, sustf, 0, True, direction)

    # if this isn't the test mass, also get the noise from the upper joint of
    # the mass below this one
    if stage_num < last_stage:
        noise += getJointTempSusceptibility(
            sus, sustf, stage_num + 1, True, direction)

    # thermal noise (m^2/Hz) for one suspension
    noise = 4 * kB * noise / w

    return noise


def suspension_thermal(f, sus, sustf):
    """Total suspension thermal noise of one suspension

    :f: frequency array in Hz
    :sus: gwinc suspension structure
    :sustf: suspension transfer function struct
    :returns: displacement noise power spectrum at :f:

    :Temp: must either be set for each stage individually, or globally
    in :sus.Temp:.  If both are specified, :sus.Temp: is interpreted as
    the temperature of the upper joint of the top stage.

    Assumes suspension transfer functions and V-H coupling have been
    pre-calculated and populated into the relevant `sus` fields.

    """
    nstage = len(sus.Stage)
    noise = np.zeros_like(f)
    for stage_num in range(nstage):
        noise += susptherm_stage(f, sus, sustf, stage_num, 'horiz')
        noise += susptherm_stage(f, sus, sustf, stage_num, 'vert')
    return abs(noise)



import pytest
import yaml

from ..struct import Struct, yaml_loader
YAMLSafeLoader = yaml_loader


TEST_YAML = '''
N: 'Test'
I:
  L: 1.3e5
  T:
    - 1
    - 2
  R:
    P: 1e-9
S:
  - M: 40
    L: 4
  - M: 30
    L: 3
'''


def validate(ifo):
    assert ifo.N == 'Test'
    assert ifo['N'] == 'Test'
    assert ifo.I.L == 1.3e5
    assert ifo['I'].L == 1.3e5
    assert ifo['I']['L'] == 1.3e5
    assert ifo['I.L'] == 1.3e5
    # assert getattr(ifo, 'I.L') == 1.3e5
    assert ifo.I.T == [1, 2]
    assert ifo['I'].T == [1, 2]
    assert ifo['I']['T'] == [1, 2]
    assert ifo['I.T'] == [1, 2]
    # assert getattr(ifo, 'I.T') == [1, 2]
    assert ifo.I.R.P == 1e-9
    assert ifo['I.R.P'] == 1e-9
    assert ifo.S[0].M == 40
    assert ifo['S[1].L'] == 3
    # assert len(ifo) == 3
    assert set(ifo) == set(['I', 'N', 'S'])
    assert set(ifo.keys()) == set(['I', 'N', 'S'])
    # assert set(ifo.values()) == []
    assert hasattr(ifo, 'N')
    assert hasattr(ifo.I, 'L')
    assert 'N' in ifo
    assert 'L' in ifo.I
    assert ifo.to_dict() == yaml.load(TEST_YAML, Loader=YAMLSafeLoader)
    assert ifo.diff(Struct.from_yaml(TEST_YAML)) == []
    with pytest.raises(AttributeError):
        ifo.A
    with pytest.raises(KeyError):
        ifo['B']


    # def basic_assert(self, ifo):
    #     assert ifo.a == 1
    #     assert ifo['a'] == 1
    #     assert ifo.b.c == 2
    #     assert ifo['b.c'] == 2
    #     assert ifo.to_dict() == self.tdict
    #     assert ifo.to_yaml() == self.tyaml
    #     # assert dict(ifo) == self.tdict
    #     assert set(ifo.keys()) == set(['a', 'b'])
    #     assert set(ifo.values()) == set([1, ifo.b])
    #     assert hasattr(ifo, 'a')
    #     assert hasattr(ifo.b, 'c')
    #     assert 'a' in ifo
    #     assert 'c' in ifo.b
    #     assert ifo.diff(Struct.from_dict(self.tdict)) == []


class TestStruct:
    def test_basic(self):
        ifo = Struct()
        ifo.N = 'Test'
        ifo.I = Struct()
        ifo.I.L = 1.3e5
        ifo['I'].T = [1, 2]
        ifo['I.R'] = Struct(dict(P=1e-9))
        ifo.S = [
            Struct(dict(M=40, L=4)),
            Struct(dict(M=30, L=3)),
        ]
        validate(ifo)

    def test_from_dict(self):
        d = yaml.load(TEST_YAML, Loader=YAMLSafeLoader)
        ifo = Struct(d)
        validate(ifo)

    def test_from_yaml(self):
        ifo = Struct.from_yaml(TEST_YAML)
        validate(ifo)

    # def test_classdef(self):
    #     class Rs(Struct):
    #         @property
    #         def P(self):
    #             return 1e-9
    #     class Is(Struct):
    #         L = 1.3e5
    #         T = [1, 2]
    #         R = Rs()
    #     class IFO(Struct):
    #         N = 'Test'
    #         I = Is()
    #         S = [
    #             Struct(dict(M=40, L=4)),
    #             Struct(dict(M=30, L=3)),
    #         ]
    #     ifo = IFO()
    #     validate(ifo)

    def test_load_override(self):
        class Is(Struct):
            @property
            def T(self):
                return [-1, -2]
        class IFO(Struct):
            N = 'Foo'
            I = Is()
        ifo = IFO.from_yaml(TEST_YAML)
        validate(ifo)

    def test_root_ref(self):
        class Ts(Struct):
            @property
            def P(self):
                return self._root * 4
        class IFO(Struct):
            T = Ts
            T.asdf = 4
            N = 'Foo'
        ifo = IFO.from_yaml(TEST_YAML)
        print(ifo.T)
        validate(ifo)

    # def test_property(self):
    #     class B(Struct):
    #         @property
    #         def c(self):
    #             return 1+1
    #     ifo = Struct()
    #     ifo.a = 1
    #     ifo.b = B()
    #     self.basic_assert(ifo)

    # def test_properties(self):
    #     ifo = Foo()
    #     with pytest.raises(AttributeError):
    #         ifo['Bar'] = 3
    #     with pytest.raises(AttributeError):
    #         ifo.Bar = 3
    #     assert ifo.Bar == 4
    #     assert 'Bar' in ifo
    #     assert hassattr(ifo, 'Bar')

    # @pytest.mark.xfail
    # def test_properties(self):
    #     ifo = Foo()
    #     assert ifo.to_dict() == dict(Bar=4)



import h5py
import yaml
import datetime

from . import logger
from . import Struct
from .trace import BudgetTrace


SCHEMA = 'GWINC noise budget'
DATA_SAVE_FORMATS = ['.hdf5', '.h5']


def _write_trace_recursive(f, trace):
    f.attrs['style'] = yaml.dump(trace.style)
    f.create_dataset('PSD', data=trace.psd)
    if trace.budget:
        bgrp = f.create_group('budget')
    for t in trace:
        tgrp = bgrp.create_group(t.name)
        _write_trace_recursive(tgrp, t)


def save_hdf5(trace, path, ifo=None, plot_style=None, **kwargs):
    """Save GWINC budget data to an HDF5 file.

    `trace` should be a BudgetTrace object (see budget.run()), and
    `path` should be the path to the output HDF5 file.  Keyword
    arguments are stored in the HDF5 top level 'attrs' key-value
    store.  If an 'ifo' keyword arg is supplied, it is assumed to be a
    Struct and will be serialized to YAML for storage.

    See HDF5_SCHEMA for information on the BudgetTrace/HDF5 storage
    format.

    """
    with h5py.File(path, 'w') as f:
        f.attrs['SCHEMA'] = SCHEMA
        f.attrs['SCHEMA_VERSION'] = 2
        # FIXME: add budget code hash or something
        f.attrs['date'] = datetime.datetime.now().isoformat()
        if ifo:
            f.attrs['ifo'] = ifo.to_yaml()
        if plot_style:
            f.attrs['plot_style'] = yaml.dump(plot_style)
        for key, val in kwargs.items():
            f.attrs[key] = val
        f.create_dataset('Freq', data=trace.freq)
        tgrp = f.create_group(trace.name)
        _write_trace_recursive(tgrp, trace)


def _load_hdf5_v1(f):
    def read_recursive(element):
        budget = []
        for name, item in element.items():
            if name == 'Total':
                total = item[:]
                continue
            if isinstance(item, h5py.Group):
                trace = read_recursive(item)
                trace.name = name
            else:
                trace = BudgetTrace(
                    name=name,
                    style=dict(item.attrs),
                    psd=item[:],
                    budget=[],
                )
            budget.append(trace)
        return BudgetTrace(
            psd=total,
            budget=budget,
        )
    trace = read_recursive(f['/traces'])
    trace.name = 'Total'
    trace._freq = f['Freq'][:]
    attrs = dict(f.attrs)
    if 'ifo' in attrs:
        try:
            ifo = Struct.from_yaml(attrs['ifo'])
            del attrs['ifo']
        except yaml.constructor.ConstructorError:
            logger.warning("HDF5 load warning: Could not de-serialize 'ifo' YAML attribute.")
            ifo = None
    trace.style = attrs
    trace.ifo = ifo
    return trace


def _load_hdf5_v2(f):
    def read_recursive(element, freq):
        psd = element['PSD'][:]
        style = yaml.safe_load(element.attrs['style'])
        budget = []
        if 'budget' in element:
            for name, item in element['budget'].items():
                trace = read_recursive(item, freq)
                trace.name = name
                budget.append(trace)
        return BudgetTrace(
            freq=freq,
            psd=psd,
            budget=budget,
            style=style,
        )
    freq = f['Freq'][:]
    for name, item in f.items():
        if name == 'Freq':
            continue
        else:
            trace = read_recursive(item, freq)
            trace.name = name
    trace.ifo = None
    attrs = dict(f.attrs)
    ifo = attrs.get('ifo')
    if ifo:
        try:
            trace.ifo = Struct.from_yaml(ifo)
            del attrs['ifo']
        except yaml.constructor.ConstructorError:
            logger.warning("HDF5 load warning: Could not de-serialize 'ifo' YAML attribute.")
    trace.plot_style = yaml.safe_load(attrs.get('plot_style', ''))
    return trace


def load_hdf5(path):
    """Load GWINC budget data from an HDF5 file.

    Returns BudgetTrace for trace data stored in the HDF5 file at
    `path`.  An 'ifo' attr will be de-serialized from YAML into a
    Struct object and assigned as an attribute to the BudgetTrace.

    See HDF5_SCHEMA for information on the BudgetTrace/HDF5 storage
    format.

    """
    loaders = {
        1: _load_hdf5_v1,
        2: _load_hdf5_v2,
    }
    with h5py.File(path, 'r') as f:
        version = f.attrs.get('SCHEMA_VERSION', 1)
        return loaders[version](f)



def plot_trace(
        trace,
        ax=None,
        **kwargs
):
    """Plot a GWINC BudgetTrace noise budget from calculated noises

    If an axis handle is provided it will be used for the plot.

    Returns the figure handle.

    """
    if ax is None:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    else:
        fig = ax.figure

    total = trace.asd
    ylim = [min(total)/10, max(total)]

    style = dict(alpha=0.6)

    style.update(getattr(trace, 'style', {}))
    if 'color' not in style and 'c' not in style:
        style['color'] = '#000000'
    if 'alpha' not in style:
        style['alpha'] = 0.6
    if 'linewidth' not in style and 'lw' not in style:
        style['linewidth'] = 4
    if 'label' in style:
        style['label'] = 'Total ' + style['label']
    else:
        style['label'] = 'Total'
    ax.loglog(trace.freq, total, **style)

    for name, strace in trace.items():
        style = strace.style
        if 'label' not in style:
            style['label'] = name
        if 'linewidth' not in style and 'lw' not in style:
            style['linewidth'] = 3
        ax.loglog(trace.freq, strace.asd, **style)

    ax.grid(
        True,
        which='major',
        linewidth=0.5,
        ls='-',
        alpha=0.5,
    )

    ax.grid(
        True,
        which='minor',
        linewidth=0.5,
        ls='-',
        alpha=0.2,
    )

    ax.legend(
        ncol=2,
        fontsize='small',
    )

    ax.autoscale(enable=True, axis='y', tight=True)
    ax.set_ylim(kwargs.get('ylim', ylim))
    ax.set_xlim(trace.freq[0], trace.freq[-1])
    ax.set_xlabel('Frequency [Hz]')
    if 'ylabel' in kwargs:
        ax.set_ylabel(kwargs['ylabel'])
    if 'title' in kwargs:
        ax.set_title(kwargs['title'])

    return fig


# FIXME: deprecate
plot_noise = plot_trace
plot_budget = plot_trace
